# Copyright 2014 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Test class for Management Interface used by iLO modules."""

import os
import shutil
from unittest import mock

import ddt
from oslo_config import cfg
from oslo_utils import importutils
from oslo_utils import uuidutils

from ironic.common import boot_devices
from ironic.common import boot_modes
from ironic.common import exception
from ironic.common import states
from ironic.conductor import task_manager
from ironic.conductor import utils as manager_utils
from ironic.drivers.modules import agent_base
from ironic.drivers.modules import deploy_utils
from ironic.drivers.modules.ilo import boot as ilo_boot
from ironic.drivers.modules.ilo import common as ilo_common
from ironic.drivers.modules.ilo import management as ilo_management
from ironic.drivers.modules import ipmitool
from ironic.drivers import utils as driver_utils
from ironic.tests.unit.db import base as db_base
from ironic.tests.unit.db import utils as db_utils
from ironic.tests.unit.drivers.modules.ilo import test_common
from ironic.tests.unit.objects import utils as obj_utils

ilo_error = importutils.try_import('proliantutils.exception')

INFO_DICT = db_utils.get_test_ilo_info()

CONF = cfg.CONF


@ddt.ddt
class IloManagementTestCase(test_common.BaseIloTest):

    def setUp(self):
        super(IloManagementTestCase, self).setUp()
        port_1 = obj_utils.create_test_port(
            self.context, node_id=self.node.id,
            address='11:22:33:44:55:66', uuid=uuidutils.generate_uuid())
        port_2 = obj_utils.create_test_port(
            self.context, node_id=self.node.id,
            address='11:22:33:44:55:67', uuid=uuidutils.generate_uuid())
        self.ports = [port_1, port_2]

    def test_get_properties(self):
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            expected = ilo_management.MANAGEMENT_PROPERTIES
            self.assertEqual(expected,
                             task.driver.management.get_properties())

    @mock.patch.object(ilo_common, 'parse_driver_info', spec_set=True,
                       autospec=True)
    def test_validate(self, driver_info_mock):
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            task.driver.management.validate(task)
            driver_info_mock.assert_called_once_with(task.node)

    def test_get_supported_boot_devices(self):
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            expected = [boot_devices.PXE, boot_devices.DISK,
                        boot_devices.CDROM, boot_devices.ISCSIBOOT]
            self.assertEqual(
                sorted(expected),
                sorted(task.driver.management.
                       get_supported_boot_devices(task)))

    @mock.patch.object(ilo_common, 'get_ilo_object', spec_set=True,
                       autospec=True)
    def test_get_boot_device_next_boot(self, get_ilo_object_mock):
        ilo_object_mock = get_ilo_object_mock.return_value
        ilo_object_mock.get_one_time_boot.return_value = 'CDROM'
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            expected_device = boot_devices.CDROM
            expected_response = {'boot_device': expected_device,
                                 'persistent': False}
            self.assertEqual(expected_response,
                             task.driver.management.get_boot_device(task))
            ilo_object_mock.get_one_time_boot.assert_called_once_with()

    @mock.patch.object(ilo_common, 'get_ilo_object', spec_set=True,
                       autospec=True)
    def test_get_boot_device_persistent(self, get_ilo_object_mock):
        ilo_mock = get_ilo_object_mock.return_value
        ilo_mock.get_one_time_boot.return_value = 'Normal'
        ilo_mock.get_persistent_boot_device.return_value = 'NETWORK'

        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            expected_device = boot_devices.PXE
            expected_response = {'boot_device': expected_device,
                                 'persistent': True}
            self.assertEqual(expected_response,
                             task.driver.management.get_boot_device(task))
            ilo_mock.get_one_time_boot.assert_called_once_with()
            ilo_mock.get_persistent_boot_device.assert_called_once_with()

    @mock.patch.object(ilo_common, 'get_ilo_object', spec_set=True,
                       autospec=True)
    def test_get_boot_device_fail(self, get_ilo_object_mock):
        ilo_mock_object = get_ilo_object_mock.return_value
        exc = ilo_error.IloError('error')
        ilo_mock_object.get_one_time_boot.side_effect = exc

        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            self.assertRaises(exception.IloOperationError,
                              task.driver.management.get_boot_device,
                              task)
        ilo_mock_object.get_one_time_boot.assert_called_once_with()

    @mock.patch.object(ilo_common, 'get_ilo_object', spec_set=True,
                       autospec=True)
    def test_get_boot_device_persistent_fail(self, get_ilo_object_mock):
        ilo_mock_object = get_ilo_object_mock.return_value
        ilo_mock_object.get_one_time_boot.return_value = 'Normal'
        exc = ilo_error.IloError('error')
        ilo_mock_object.get_persistent_boot_device.side_effect = exc

        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            self.assertRaises(exception.IloOperationError,
                              task.driver.management.get_boot_device,
                              task)
        ilo_mock_object.get_one_time_boot.assert_called_once_with()
        ilo_mock_object.get_persistent_boot_device.assert_called_once_with()

    @mock.patch.object(ilo_common, 'get_ilo_object', spec_set=True,
                       autospec=True)
    def test_set_boot_device_ok(self, get_ilo_object_mock):
        ilo_object_mock = get_ilo_object_mock.return_value
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            task.driver.management.set_boot_device(task, boot_devices.CDROM,
                                                   False)
            get_ilo_object_mock.assert_called_once_with(task.node)
            ilo_object_mock.set_one_time_boot.assert_called_once_with('CDROM')

    @mock.patch.object(ilo_common, 'get_ilo_object', spec_set=True,
                       autospec=True)
    def test_set_boot_device_persistent_true(self, get_ilo_object_mock):
        ilo_mock = get_ilo_object_mock.return_value
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            task.driver.management.set_boot_device(task, boot_devices.PXE,
                                                   True)
            get_ilo_object_mock.assert_called_once_with(task.node)
            ilo_mock.update_persistent_boot.assert_called_once_with(
                ['NETWORK'])

    @mock.patch.object(ilo_common, 'get_ilo_object', spec_set=True,
                       autospec=True)
    def test_set_boot_device_fail(self, get_ilo_object_mock):
        ilo_mock_object = get_ilo_object_mock.return_value
        exc = ilo_error.IloError('error')
        ilo_mock_object.set_one_time_boot.side_effect = exc

        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            self.assertRaises(exception.IloOperationError,
                              task.driver.management.set_boot_device,
                              task, boot_devices.PXE)
        ilo_mock_object.set_one_time_boot.assert_called_once_with('NETWORK')

    @mock.patch.object(ilo_common, 'get_ilo_object', spec_set=True,
                       autospec=True)
    def test_set_boot_device_persistent_fail(self, get_ilo_object_mock):
        ilo_mock_object = get_ilo_object_mock.return_value
        exc = ilo_error.IloError('error')
        ilo_mock_object.update_persistent_boot.side_effect = exc

        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            self.assertRaises(exception.IloOperationError,
                              task.driver.management.set_boot_device,
                              task, boot_devices.PXE, True)
        ilo_mock_object.update_persistent_boot.assert_called_once_with(
            ['NETWORK'])

    def test_set_boot_device_invalid_device(self):
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            self.assertRaises(exception.InvalidParameterValue,
                              task.driver.management.set_boot_device,
                              task, 'fake-device')

    @mock.patch.object(ilo_common, 'update_ipmi_properties', spec_set=True,
                       autospec=True)
    @mock.patch.object(ipmitool.IPMIManagement, 'get_sensors_data',
                       spec_set=True, autospec=True)
    def test_get_sensor_data(self, get_sensors_data_mock, update_ipmi_mock):
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            task.driver.management.get_sensors_data(task)
            update_ipmi_mock.assert_called_once_with(task)
            get_sensors_data_mock.assert_called_once_with(mock.ANY, task)

    @mock.patch.object(ilo_common, 'get_ilo_object', spec_set=True,
                       autospec=True)
    def test__execute_ilo_step_ok(self, get_ilo_object_mock):
        ilo_mock = get_ilo_object_mock.return_value
        step_mock = getattr(ilo_mock, 'fake-step')
        ilo_management._execute_ilo_step(
            self.node, 'fake-step', 'args', kwarg='kwarg')
        step_mock.assert_called_once_with('args', kwarg='kwarg')

    @mock.patch.object(ilo_management, 'LOG', spec_set=True, autospec=True)
    @mock.patch.object(ilo_common, 'get_ilo_object', spec_set=True,
                       autospec=True)
    def test__execute_ilo_step_not_supported(self, get_ilo_object_mock,
                                             log_mock):
        ilo_mock = get_ilo_object_mock.return_value
        exc = ilo_error.IloCommandNotSupportedError("error")
        step_mock = getattr(ilo_mock, 'fake-step')
        step_mock.side_effect = exc
        ilo_management._execute_ilo_step(
            self.node, 'fake-step', 'args', kwarg='kwarg')
        step_mock.assert_called_once_with('args', kwarg='kwarg')
        self.assertTrue(log_mock.warning.called)

    @mock.patch.object(ilo_common, 'get_ilo_object', spec_set=True,
                       autospec=True)
    def _test__execute_ilo_step_fail(self, get_ilo_object_mock):
        if self.node.clean_step:
            step = self.node.clean_step
            step_name = step['step']
            exept = exception.NodeCleaningFailure
        else:
            step = self.node.deploy_step
            step_name = step['step']
            exept = exception.InstanceDeployFailure
        ilo_mock = get_ilo_object_mock.return_value
        exc = ilo_error.IloError("error")
        step_mock = getattr(ilo_mock, step_name)
        step_mock.side_effect = exc
        self.assertRaises(exept,
                          ilo_management._execute_ilo_step,
                          self.node, step_name, 'args', kwarg='kwarg')
        step_mock.assert_called_once_with('args', kwarg='kwarg')

    def test__execute_ilo_step_fail_clean(self):
        self.node.clean_step = {'priority': 100, 'interface': 'management',
                                'step': 'fake-step',
                                'argsinfo': {}}
        self.node.save()
        self._test__execute_ilo_step_fail()

    def test__execute_ilo_step_fail_deploy(self):
        self.node.deploy_step = {'priority': 100, 'interface': 'management',
                                 'step': 'fake-step',
                                 'argsinfo': {}}
        self.node.save()
        self._test__execute_ilo_step_fail()

    @mock.patch.object(deploy_utils, 'build_agent_options',
                       spec_set=True, autospec=True)
    @mock.patch.object(ilo_boot.IloVirtualMediaBoot, 'clean_up_ramdisk',
                       spec_set=True, autospec=True)
    @mock.patch.object(ilo_boot.IloVirtualMediaBoot, 'prepare_ramdisk',
                       spec_set=True, autospec=True)
    @mock.patch.object(ilo_management, '_execute_ilo_step',
                       spec_set=True, autospec=True)
    def test_reset_ilo(
            self, execute_step_mock, prepare_mock, cleanup_mock, build_mock):
        build_mock.return_value = {'a': 'b'}
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            task.driver.management.reset_ilo(task)
            execute_step_mock.assert_called_once_with(task.node, 'reset_ilo')
            cleanup_mock.assert_called_once_with(mock.ANY, task)
            build_mock.assert_called_once_with(task.node)
            prepare_mock.assert_called_once_with(
                mock.ANY, mock.ANY, {'a': 'b'})

    @mock.patch.object(ilo_management, '_execute_ilo_step',
                       spec_set=True, autospec=True)
    def test_reset_ilo_credential_ok(self, step_mock):
        info = self.node.driver_info
        info['ilo_change_password'] = "fake-password"
        self.node.driver_info = info
        self.node.save()
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            task.driver.management.reset_ilo_credential(task)
            step_mock.assert_called_once_with(
                task.node, 'reset_ilo_credential', 'fake-password')
            self.assertNotIn('ilo_change_password', task.node.driver_info)
            self.assertEqual('fake-password',
                             task.node.driver_info['ilo_password'])

    @mock.patch.object(ilo_management, '_execute_ilo_step',
                       spec_set=True, autospec=True)
    def test_reset_ilo_credential_pass_as_arg_ok(self, step_mock):
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            task.driver.management.reset_ilo_credential(
                task, change_password='fake-password')
            step_mock.assert_called_once_with(
                task.node, 'reset_ilo_credential', 'fake-password')
            self.assertNotIn('ilo_change_password', task.node.driver_info)
            self.assertEqual('fake-password',
                             task.node.driver_info['ilo_password'])

    @mock.patch.object(ilo_management, 'LOG', spec_set=True, autospec=True)
    @mock.patch.object(ilo_management, '_execute_ilo_step',
                       spec_set=True, autospec=True)
    def test_reset_ilo_credential_no_password(self, step_mock,
                                              log_mock):
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            task.driver.management.reset_ilo_credential(task)
            self.assertFalse(step_mock.called)
            self.assertTrue(log_mock.info.called)

    @mock.patch.object(ilo_management, '_execute_ilo_step',
                       spec_set=True, autospec=True)
    def test_reset_bios_to_default(self, step_mock):
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            task.driver.management.reset_bios_to_default(task)
            step_mock.assert_called_once_with(task.node,
                                              'reset_bios_to_default')

    @mock.patch.object(ilo_management, '_execute_ilo_step',
                       spec_set=True, autospec=True)
    def test_reset_secure_boot_keys_to_default(self, step_mock):
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            task.driver.management.reset_secure_boot_keys_to_default(task)
            step_mock.assert_called_once_with(task.node,
                                              'reset_secure_boot_keys')

    @mock.patch.object(ilo_management, '_execute_ilo_step',
                       spec_set=True, autospec=True)
    def test_clear_secure_boot_keys(self, step_mock):
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            task.driver.management.clear_secure_boot_keys(task)
            step_mock.assert_called_once_with(task.node,
                                              'clear_secure_boot_keys')

    @mock.patch.object(ilo_management, '_execute_ilo_step',
                       spec_set=True, autospec=True)
    def test_activate_license(self, step_mock):
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            activate_license_args = {
                'ilo_license_key': 'XXXXX-YYYYY-ZZZZZ-XYZZZ-XXYYZ'}
            task.driver.management.activate_license(task,
                                                    **activate_license_args)
            step_mock.assert_called_once_with(
                task.node, 'activate_license', 'XXXXX-YYYYY-ZZZZZ-XYZZZ-XXYYZ')

    @mock.patch.object(ilo_management, '_execute_ilo_step',
                       spec_set=True, autospec=True)
    def test_activate_license_no_or_invalid_format_license_key(
            self, step_mock):
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            for license_key_value in (None, [], {}):
                activate_license_args = {'ilo_license_key': license_key_value}
                self.assertRaises(exception.InvalidParameterValue,
                                  task.driver.management.activate_license,
                                  task,
                                  **activate_license_args)
                self.assertFalse(step_mock.called)

    @mock.patch.object(ilo_management, '_execute_ilo_step',
                       spec_set=True, autospec=True)
    def test_security_parameters_update(self, step_mock):
        security_parameters = [
            {
                "param": "password_complexity",
                "enable": True,
                "ignore": False
            }
        ]
        security_parameters_args = {
            'security_parameters': security_parameters}

        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            task.driver.management.security_parameters_update(
                task, **security_parameters_args)
            step_mock.assert_called_once_with(
                task.node, 'update_password_complexity', True, False)

    @mock.patch.object(ilo_management, '_execute_ilo_step',
                       spec_set=True, autospec=True)
    def test_update_minimum_password_length(self, step_mock):
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            minimum_password_args = {'password_length': '8'}
            task.driver.management.update_minimum_password_length(
                task, **minimum_password_args)
            step_mock.assert_called_once_with(
                task.node, 'update_minimum_password_length', '8', False)

    @mock.patch.object(ilo_management, '_execute_ilo_step',
                       spec_set=True, autospec=True)
    def test_update_auth_failure_logging_threshold(self, step_mock):
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            auth_failure_args = {'logging_threshold': '1'}
            task.driver.management.update_auth_failure_logging_threshold(
                task, **auth_failure_args)
            step_mock.assert_called_once_with(
                task.node, 'update_authentication_failure_logging', '1', False)

    @mock.patch.object(ilo_management, '_execute_ilo_step',
                       spec_set=True, autospec=True)
    @mock.patch.object(os, 'makedirs', spec_set=True, autospec=True)
    def test_create_csr(self, os_mock, step_mock):
        csr_params_args = {
            "City": "Bangalore",
            "CommonName": "1.1.1.1",
            "Country": "ABC",
            "OrgName": "DEF",
            "State": "IJK"
        }
        csr_args = {
            "csr_params": csr_params_args}
        CONF.ilo.cert_path = "/var/lib/ironic/ilo"

        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            task.driver.management.create_csr(task, **csr_args)
            cert_path = os.path.join(CONF.ilo.cert_path, self.node.uuid)
            step_mock.assert_called_once_with(task.node, 'create_csr',
                                              cert_path, csr_params_args)
            os_mock.assert_called_once_with(cert_path, 0o755)

    @mock.patch.object(ilo_management, '_execute_ilo_step',
                       spec_set=True, autospec=True)
    @mock.patch.object(os, 'makedirs', spec_set=True, autospec=True)
    @mock.patch.object(shutil, 'copy', spec_set=True, autospec=True)
    def test_add_https_certificate(self, shutil_mock, os_mock,
                                   step_mock):
        CONF.ilo.cert_path = "/var/lib/ironic/ilo"
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            cert_file_args = {'cert_file': '/test1/cert'}
            task.driver.management.add_https_certificate(
                task, **cert_file_args)
            cert_path = os.path.join(CONF.ilo.cert_path, self.node.uuid)
            cert_path_name = os.path.join(cert_path, self.node.uuid)
            filename = cert_path_name + ".crt"
            step_mock.assert_called_once_with(
                task.node, 'add_https_certificate', filename)
            os_mock.assert_called_once_with(cert_path, 0o755)
            shutil_mock.assert_called_once_with('/test1/cert', filename)

    @mock.patch.object(ilo_management, '_execute_ilo_step',
                       spec_set=True, autospec=True)
    @mock.patch.object(os, 'makedirs', spec_set=True, autospec=True)
    @mock.patch.object(shutil, 'copy', spec_set=True, autospec=True)
    @mock.patch.object(ilo_common, 'download', spec_set=True, autospec=True)
    def test_add_https_certificate_fileurl(self, download_mock, shutil_mock,
                                           os_mock, step_mock):
        CONF.ilo.cert_path = "/var/lib/ironic/ilo"
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            cert_file_args = {'cert_file': 'file:///test1/cert'}
            task.driver.management.add_https_certificate(
                task, **cert_file_args)
            cert_path = os.path.join(CONF.ilo.cert_path, self.node.uuid)
            cert_path_name = os.path.join(cert_path, self.node.uuid)
            fname = cert_path_name + ".crt"
            step_mock.assert_called_once_with(
                task.node, 'add_https_certificate', fname)
            os_mock.assert_called_once_with(cert_path, 0o755)
            shutil_mock.assert_not_called()
            download_mock.assert_called_once_with(fname, 'file:///test1/cert')

    @mock.patch.object(ilo_management, '_execute_ilo_step',
                       spec_set=True, autospec=True)
    @mock.patch.object(os, 'makedirs', spec_set=True, autospec=True)
    @mock.patch.object(shutil, 'copy', spec_set=True, autospec=True)
    @mock.patch.object(ilo_common, 'download', spec_set=True, autospec=True)
    def test_add_https_certificate_httpurl(self, download_mock, shutil_mock,
                                           os_mock, step_mock):
        CONF.ilo.cert_path = "/var/lib/ironic/ilo"
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            cert_file_args = {'cert_file': 'http://1.1.1.1/cert'}
            task.driver.management.add_https_certificate(
                task, **cert_file_args)
            cert_path = os.path.join(CONF.ilo.cert_path, self.node.uuid)
            cert_path_name = os.path.join(cert_path, self.node.uuid)
            fname = cert_path_name + ".crt"
            step_mock.assert_called_once_with(
                task.node, 'add_https_certificate', fname)
            os_mock.assert_called_once_with(cert_path, 0o755)
            shutil_mock.assert_not_called()
            download_mock.assert_called_once_with(fname, 'http://1.1.1.1/cert')

    @mock.patch.object(ilo_management, '_execute_ilo_step',
                       spec_set=True, autospec=True)
    @mock.patch.object(os, 'makedirs', spec_set=True, autospec=True)
    @mock.patch.object(shutil, 'copy', spec_set=True, autospec=True)
    @mock.patch.object(ilo_common, 'download', spec_set=True, autospec=True)
    def test_add_https_certificate_url_exception(self, download_mock,
                                                 shutil_mock, os_mock,
                                                 step_mock):
        CONF.ilo.cert_path = "/var/lib/ironic/ilo"
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            cert_file_args = {'cert_file': 'swift://1.1.1.1/cert'}
            self.assertRaises(exception.IloOperationNotSupported,
                              task.driver.management.add_https_certificate,
                              task,
                              **cert_file_args)

            cert_path = os.path.join(CONF.ilo.cert_path, self.node.uuid)
            step_mock.assert_not_called()
            os_mock.assert_called_once_with(cert_path, 0o755)
            shutil_mock.assert_not_called()
            download_mock.assert_not_called()

    @mock.patch.object(deploy_utils, 'build_agent_options',
                       spec_set=True, autospec=True)
    @mock.patch.object(ilo_boot.IloVirtualMediaBoot, 'clean_up_ramdisk',
                       spec_set=True, autospec=True)
    @mock.patch.object(ilo_boot.IloVirtualMediaBoot, 'prepare_ramdisk',
                       spec_set=True, autospec=True)
    @mock.patch.object(ilo_management, 'LOG', autospec=True)
    @mock.patch.object(ilo_management, '_execute_ilo_step',
                       spec_set=True, autospec=True)
    @mock.patch.object(ilo_management.firmware_processor, 'FirmwareProcessor',
                       spec_set=True, autospec=True)
    @mock.patch.object(ilo_common, 'remove_single_or_list_of_files',
                       spec_set=True, autospec=True)
    def _test_update_firmware_calls_step_foreach_url(
            self, remove_file_mock, FirmwareProcessor_mock, execute_step_mock,
            LOG_mock, prepare_mock, cleanup_mock, build_mock):
        if self.node.clean_step:
            step = self.node.clean_step
        else:
            step = self.node.deploy_step
        build_mock.return_value = {'a': 'b'}
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            firmware_update_args = step['argsinfo']
            FirmwareProcessor_mock.return_value.process_fw_on.side_effect = [
                ilo_management.firmware_processor.FirmwareImageLocation(
                    'fw_location_for_filepath', 'filepath'),
                ilo_management.firmware_processor.FirmwareImageLocation(
                    'fw_location_for_httppath', 'httppath'),
                ilo_management.firmware_processor.FirmwareImageLocation(
                    'fw_location_for_httpspath', 'httpspath'),
                ilo_management.firmware_processor.FirmwareImageLocation(
                    'fw_location_for_swiftpath', 'swiftpath'),
                ilo_management.firmware_processor.FirmwareImageLocation(
                    'fw_location_for_another_filepath', 'filepath2')
            ]
            task.driver.management.update_firmware(task,
                                                   **firmware_update_args)
            calls = [mock.call(task.node, 'update_firmware',
                               'fw_location_for_filepath', 'ilo'),
                     mock.call(task.node, 'update_firmware',
                               'fw_location_for_httppath', 'cpld'),
                     mock.call(task.node, 'update_firmware',
                               'fw_location_for_httpspath', 'power_pic'),
                     mock.call(task.node, 'update_firmware',
                               'fw_location_for_swiftpath', 'bios'),
                     mock.call(task.node, 'update_firmware',
                               'fw_location_for_another_filepath', 'chassis'),
                     ]
            execute_step_mock.assert_has_calls(calls)
            self.assertEqual(5, execute_step_mock.call_count)
            cleanup_mock.assert_called_once_with(mock.ANY, task)
            build_mock.assert_called_once_with(task.node)
            prepare_mock.assert_called_once_with(
                mock.ANY, mock.ANY, {'a': 'b'})

    def test_update_firmware_calls_step_foreach_url_clean(self):
        firmware_images = [
            {
                'url': 'file:///any_path',
                'checksum': 'xxxx',
                'component': 'ilo'
            },
            {
                'url': 'http://any_url',
                'checksum': 'xxxx',
                'component': 'cpld'
            },
            {
                'url': 'https://any_url',
                'checksum': 'xxxx',
                'component': 'power_pic'
            },
            {
                'url': 'swift://container/object',
                'checksum': 'xxxx',
                'component': 'bios'
            },
            {
                'url': 'file:///any_path',
                'checksum': 'xxxx',
                'component': 'chassis'
            }
        ]
        firmware_update_args = {'firmware_update_mode': 'ilo',
                                'firmware_images': firmware_images}
        self.node.clean_step = {'priority': 100, 'interface': 'management',
                                'step': 'update_firmware',
                                'argsinfo': firmware_update_args}
        self.node.save()
        self._test_update_firmware_calls_step_foreach_url()

    def test_update_firmware_calls_step_foreach_url_deploy(self):
        firmware_images = [
            {
                'url': 'file:///any_path',
                'checksum': 'xxxx',
                'component': 'ilo'
            },
            {
                'url': 'http://any_url',
                'checksum': 'xxxx',
                'component': 'cpld'
            },
            {
                'url': 'https://any_url',
                'checksum': 'xxxx',
                'component': 'power_pic'
            },
            {
                'url': 'swift://container/object',
                'checksum': 'xxxx',
                'component': 'bios'
            },
            {
                'url': 'file:///any_path',
                'checksum': 'xxxx',
                'component': 'chassis'
            }
        ]
        firmware_update_args = {'firmware_update_mode': 'ilo',
                                'firmware_images': firmware_images}
        self.node.deploy_step = {'priority': 100, 'interface': 'management',
                                 'step': 'update_firmware',
                                 'argsinfo': firmware_update_args}
        self.node.save()
        self._test_update_firmware_calls_step_foreach_url()

    def _test_update_firmware_invalid_update_mode_provided(self):
        if self.node.clean_step:
            step = self.node.clean_step
        else:
            step = self.node.deploy_step
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            firmware_update_args = step['argsinfo']
            firmware_update_args = {'firmware_update_mode': 'invalid_mode',
                                    'firmware_images': None}
            self.assertRaises(exception.InvalidParameterValue,
                              task.driver.management.update_firmware,
                              task,
                              **firmware_update_args)

    def test_update_firmware_invalid_update_mode_provided_clean(self):
        firmware_update_args = {'firmware_update_mode': 'invalid_mode',
                                'firmware_images': None}
        self.node.clean_step = {'priority': 100, 'interface': 'management',
                                'step': 'update_firmware',
                                'argsinfo': firmware_update_args}
        self.node.save()
        self._test_update_firmware_invalid_update_mode_provided()

    def test_update_firmware_invalid_update_mode_provided_deploy(self):
        firmware_update_args = {'firmware_update_mode': 'invalid_mode',
                                'firmware_images': None}
        self.node.deploy_step = {'priority': 100, 'interface': 'management',
                                 'step': 'update_firmware',
                                 'argsinfo': firmware_update_args}
        self.node.save()
        self._test_update_firmware_invalid_update_mode_provided()

    def _test_update_firmware_error_for_no_firmware_url(self):
        if self.node.clean_step:
            step = self.node.clean_step
        else:
            step = self.node.deploy_step
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            firmware_update_args = step['argsinfo']
            firmware_update_args = {'firmware_update_mode': 'ilo',
                                    'firmware_images': []}
            self.assertRaises(exception.InvalidParameterValue,
                              task.driver.management.update_firmware,
                              task,
                              **firmware_update_args)

    def test_update_firmware_error_for_no_firmware_url_clean(self):
        firmware_update_args = {'firmware_update_mode': 'ilo',
                                'firmware_images': []}
        self.node.clean_step = {'priority': 100, 'interface': 'management',
                                'step': 'update_firmware',
                                'argsinfo': firmware_update_args}
        self.node.save()
        self._test_update_firmware_error_for_no_firmware_url()

    def test_update_firmware_error_for_no_firmware_url_deploy(self):
        firmware_update_args = {'firmware_update_mode': 'ilo',
                                'firmware_images': []}
        self.node.deploy_step = {'priority': 100, 'interface': 'management',
                                 'step': 'update_firmware',
                                 'argsinfo': firmware_update_args}
        self.node.save()
        self._test_update_firmware_error_for_no_firmware_url()

    def _test_update_firmware_throws_error_for_invalid_component_type(self):
        if self.node.clean_step:
            step = self.node.clean_step
            exept = exception.NodeCleaningFailure
        else:
            step = self.node.deploy_step
            exept = exception.InstanceDeployFailure
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            firmware_update_args = step['argsinfo']
            self.assertRaises(exept,
                              task.driver.management.update_firmware,
                              task,
                              **firmware_update_args)

    def test_update_firmware_error_for_invalid_component_type_clean(self):
        firmware_update_args = {'firmware_update_mode': 'ilo',
                                'firmware_images': [
                                    {
                                        'url': 'any_valid_url',
                                        'checksum': 'xxxx',
                                        'component': 'xyz'
                                    }
                                ]}
        self.node.clean_step = {'priority': 100, 'interface': 'management',
                                'step': 'update_firmware',
                                'argsinfo': firmware_update_args}
        self.node.save()
        self._test_update_firmware_throws_error_for_invalid_component_type()

    def test_update_firmware_error_for_invalid_component_type_deploy(self):
        firmware_update_args = {'firmware_update_mode': 'ilo',
                                'firmware_images': [
                                    {
                                        'url': 'any_valid_url',
                                        'checksum': 'xxxx',
                                        'component': 'xyz'
                                    }
                                ]}
        self.node.deploy_step = {'priority': 100, 'interface': 'management',
                                 'step': 'update_firmware',
                                 'argsinfo': firmware_update_args}
        self.node.save()
        self._test_update_firmware_throws_error_for_invalid_component_type()

    @mock.patch.object(ilo_management, 'LOG', autospec=True)
    @mock.patch.object(ilo_management.firmware_processor.FirmwareProcessor,
                       'process_fw_on', spec_set=True, autospec=True)
    def _test_update_firmware_throws_error_for_checksum_validation_error(
            self, process_fw_on_mock, LOG_mock):
        if self.node.clean_step:
            step = self.node.clean_step
            exept = exception.NodeCleaningFailure
        else:
            step = self.node.deploy_step
            exept = exception.InstanceDeployFailure
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            # | GIVEN |
            firmware_update_args = step['argsinfo']
            process_fw_on_mock.side_effect = exception.ImageRefValidationFailed
            # | WHEN & THEN |
            self.assertRaises(exept,
                              task.driver.management.update_firmware,
                              task,
                              **firmware_update_args)

    def test_update_firmware_error_for_checksum_validation_error_clean(self):
        firmware_update_args = {'firmware_update_mode': 'ilo',
                                'firmware_images': [
                                    {
                                        'url': 'any_valid_url',
                                        'checksum': 'invalid_checksum',
                                        'component': 'bios'
                                    }
                                ]}
        self.node.clean_step = {'priority': 100, 'interface': 'management',
                                'step': 'update_firmware',
                                'argsinfo': firmware_update_args}
        self.node.save()
        self._test_update_firmware_throws_error_for_checksum_validation_error()

    def test_update_firmware_error_for_checksum_validation_error_deploy(self):
        firmware_update_args = {'firmware_update_mode': 'ilo',
                                'firmware_images': [
                                    {
                                        'url': 'any_valid_url',
                                        'checksum': 'invalid_checksum',
                                        'component': 'bios'
                                    }
                                ]}
        self.node.deploy_step = {'priority': 100, 'interface': 'management',
                                 'step': 'update_firmware',
                                 'argsinfo': firmware_update_args}
        self.node.save()
        self._test_update_firmware_throws_error_for_checksum_validation_error()

    @mock.patch.object(ilo_management, '_execute_ilo_step',
                       spec_set=True, autospec=True)
    @mock.patch.object(ilo_management.firmware_processor, 'FirmwareProcessor',
                       spec_set=True, autospec=True)
    def _test_update_firmware_doesnt_update_any_if_any_url_fails(
            self, FirmwareProcessor_mock, clean_step_mock):
        """update_firmware throws error for failure in processing any url

        update_firmware doesn't invoke firmware update of proliantutils
        for any url if processing on any firmware url fails.
        """
        if self.node.clean_step:
            step = self.node.clean_step
            exept = exception.NodeCleaningFailure
        else:
            step = self.node.deploy_step
            exept = exception.InstanceDeployFailure
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            firmware_update_args = step['argsinfo']
            FirmwareProcessor_mock.return_value.process_fw_on.side_effect = [
                ilo_management.firmware_processor.FirmwareImageLocation(
                    'extracted_firmware_url_of_any_valid_url', 'filename'),
                exception.IronicException
            ]
            self.assertRaises(exept,
                              task.driver.management.update_firmware,
                              task,
                              **firmware_update_args)
            self.assertFalse(clean_step_mock.called)

    def test_update_firmware_doesnt_update_any_if_any_url_fails_clean(self):
        firmware_update_args = {'firmware_update_mode': 'ilo',
                                'firmware_images': [
                                    {
                                        'url': 'any_valid_url',
                                        'checksum': 'xxxx',
                                        'component': 'ilo'
                                    },
                                    {
                                        'url': 'any_invalid_url',
                                        'checksum': 'xxxx',
                                        'component': 'bios'
                                    }]
                                }
        self.node.clean_step = {'priority': 100, 'interface': 'management',
                                'step': 'update_firmware',
                                'argsinfo': firmware_update_args}
        self.node.save()
        self._test_update_firmware_doesnt_update_any_if_any_url_fails()

    def test_update_firmware_doesnt_update_any_if_any_url_fails_deploy(self):
        firmware_update_args = {'firmware_update_mode': 'ilo',
                                'firmware_images': [
                                    {
                                        'url': 'any_valid_url',
                                        'checksum': 'xxxx',
                                        'component': 'ilo'
                                    },
                                    {
                                        'url': 'any_invalid_url',
                                        'checksum': 'xxxx',
                                        'component': 'bios'
                                    }]
                                }
        self.node.deploy_step = {'priority': 100, 'interface': 'management',
                                 'step': 'update_firmware',
                                 'argsinfo': firmware_update_args}
        self.node.save()
        self._test_update_firmware_doesnt_update_any_if_any_url_fails()

    @mock.patch.object(ilo_management, 'LOG', autospec=True)
    @mock.patch.object(ilo_management, '_execute_ilo_step',
                       spec_set=True, autospec=True)
    @mock.patch.object(ilo_management.firmware_processor, 'FirmwareProcessor',
                       spec_set=True, autospec=True)
    @mock.patch.object(ilo_management.firmware_processor.FirmwareImageLocation,
                       'remove', spec_set=True, autospec=True)
    def _test_update_firmware_cleans_all_files_if_exc_thrown(
            self, remove_mock, FirmwareProcessor_mock, clean_step_mock,
            LOG_mock):
        if self.node.clean_step:
            step = self.node.clean_step
            exept = exception.NodeCleaningFailure
        else:
            step = self.node.deploy_step
            exept = exception.InstanceDeployFailure
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            firmware_update_args = step['argsinfo']
            fw_loc_obj_1 = (ilo_management.firmware_processor.
                            FirmwareImageLocation('extracted_firmware_url_1',
                                                  'filename_1'))
            fw_loc_obj_2 = (ilo_management.firmware_processor.
                            FirmwareImageLocation('extracted_firmware_url_2',
                                                  'filename_2'))
            FirmwareProcessor_mock.return_value.process_fw_on.side_effect = [
                fw_loc_obj_1, fw_loc_obj_2
            ]
            clean_step_mock.side_effect = exept(
                node=self.node.uuid, reason='ilo_exc')
            self.assertRaises(exept,
                              task.driver.management.update_firmware,
                              task,
                              **firmware_update_args)
            clean_step_mock.assert_called_once_with(
                task.node, 'update_firmware',
                'extracted_firmware_url_1', 'ilo')
            self.assertTrue(LOG_mock.error.called)
            remove_mock.assert_has_calls([mock.call(fw_loc_obj_1),
                                          mock.call(fw_loc_obj_2)])

    def test_update_firmware_cleans_all_files_if_exc_thrown_clean(self):
        firmware_update_args = {'firmware_update_mode': 'ilo',
                                'firmware_images': [
                                    {
                                        'url': 'any_valid_url',
                                        'checksum': 'xxxx',
                                        'component': 'ilo'
                                    },
                                    {
                                        'url': 'any_invalid_url',
                                        'checksum': 'xxxx',
                                        'component': 'bios'
                                    }]
                                }
        self.node.clean_step = {'priority': 100, 'interface': 'management',
                                'step': 'update_firmware',
                                'argsinfo': firmware_update_args}
        self.node.save()
        self._test_update_firmware_cleans_all_files_if_exc_thrown()

    def test_update_firmware_cleans_all_files_if_exc_thrown_deploy(self):
        firmware_update_args = {'firmware_update_mode': 'ilo',
                                'firmware_images': [
                                    {
                                        'url': 'any_valid_url',
                                        'checksum': 'xxxx',
                                        'component': 'ilo'
                                    },
                                    {
                                        'url': 'any_invalid_url',
                                        'checksum': 'xxxx',
                                        'component': 'bios'
                                    }]
                                }
        self.node.deploy_step = {'priority': 100, 'interface': 'management',
                                 'step': 'update_firmware',
                                 'argsinfo': firmware_update_args}
        self.node.save()
        self._test_update_firmware_cleans_all_files_if_exc_thrown()

    @mock.patch.object(ilo_common, 'attach_vmedia', spec_set=True,
                       autospec=True)
    @mock.patch.object(agent_base, 'execute_step', autospec=True)
    def _test_write_firmware_sum_mode_with_component(
            self, execute_mock, attach_vmedia_mock, step_type='clean'):
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            # | GIVEN |
            firmware_update_args = {
                'url': 'http://any_url',
                'checksum': 'xxxx',
                'component': ['CP02345.scexe', 'CP02567.exe']}
            step = {'interface': 'management',
                    'args': firmware_update_args}
            if step_type == 'clean':
                step['step'] = 'update_firmware_sum'
                task.node.provision_state = states.CLEANING
                execute_mock.return_value = states.CLEANWAIT
                task.node.clean_step = step
                func = task.driver.management.update_firmware_sum
                exp_ret_state = states.CLEANWAIT
            else:
                step['step'] = 'flash_firmware_sum'
                task.node.provision_state = states.DEPLOYING
                execute_mock.return_value = states.DEPLOYWAIT
                task.node.deploy_step = step
                func = task.driver.management.flash_firmware_sum
                exp_ret_state = states.DEPLOYWAIT
            # | WHEN |
            return_value = func(task, **firmware_update_args)
            # | THEN |
            attach_vmedia_mock.assert_any_call(
                task.node, 'CDROM', 'http://any_url')
            self.assertEqual(exp_ret_state, return_value)
            execute_mock.assert_called_once_with(task, step, step_type)

    def test_update_firmware_sum_mode_with_component(self):
        self._test_write_firmware_sum_mode_with_component(step_type='clean')

    def test_flash_firmware_sum_mode_with_component(self):
        self._test_write_firmware_sum_mode_with_component(step_type='deploy')

    @mock.patch.object(ilo_common, 'attach_vmedia', spec_set=True,
                       autospec=True)
    @mock.patch.object(ilo_management.firmware_processor,
                       'get_swift_url', autospec=True)
    @mock.patch.object(agent_base, 'execute_step', autospec=True)
    def _test_write_firmware_sum_mode_swift_url(
            self, execute_mock, swift_url_mock, attach_vmedia_mock,
            step_type='clean'):
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            # | GIVEN |
            swift_url_mock.return_value = "http://path-to-file"
            firmware_update_args = {
                'url': 'swift://container/object',
                'checksum': 'xxxx',
                'components': ['CP02345.scexe', 'CP02567.exe']}
            step = {'interface': 'management',
                    'args': firmware_update_args}
            if step_type == 'clean':
                task.node.provision_state = states.CLEANING
                execute_mock.return_value = states.CLEANWAIT
                step['step'] = 'update_firmware_sum',
                task.node.clean_step = step
                func = task.driver.management.update_firmware_sum
                exp_ret_state = states.CLEANWAIT
                args_data = task.node.clean_step['args']
            else:
                task.node.provision_state = states.DEPLOYING
                execute_mock.return_value = states.DEPLOYWAIT
                step['step'] = 'flash_firmware_sum',
                task.node.deploy_step = step
                func = task.driver.management.flash_firmware_sum
                exp_ret_state = states.DEPLOYWAIT
                args_data = task.node.deploy_step['args']
            # | WHEN |
            return_value = func(task, **firmware_update_args)
            # | THEN |
            attach_vmedia_mock.assert_any_call(
                task.node, 'CDROM', 'http://path-to-file')
            self.assertEqual(exp_ret_state, return_value)
            self.assertEqual(args_data['url'], "http://path-to-file")

    def test_write_firmware_sum_mode_swift_url_clean(self):
        self._test_write_firmware_sum_mode_swift_url(step_type='clean')

    def test_write_firmware_sum_mode_swift_url_deploy(self):
        self._test_write_firmware_sum_mode_swift_url(step_type='deploy')

    @mock.patch.object(ilo_common, 'attach_vmedia', spec_set=True,
                       autospec=True)
    @mock.patch.object(agent_base, 'execute_step', autospec=True)
    def _test_write_firmware_sum_mode_without_component(
            self, execute_mock, attach_vmedia_mock, step_type='clean'):
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            # | GIVEN |
            firmware_update_args = {
                'url': 'any_valid_url',
                'checksum': 'xxxx'}
            step = {'interface': 'management',
                    'args': firmware_update_args}
            if step_type == 'clean':
                task.node.provision_state = states.CLEANING
                execute_mock.return_value = states.CLEANWAIT
                step['step'] = 'update_firmware_sum'
                task.node.clean_step = step
                func = task.driver.management.update_firmware_sum
                exp_ret_state = states.CLEANWAIT
            else:
                task.node.provision_state = states.DEPLOYING
                execute_mock.return_value = states.DEPLOYWAIT
                step['step'] = 'flash_firmware_sum'
                task.node.deploy_step = step
                func = task.driver.management.flash_firmware_sum
                exp_ret_state = states.DEPLOYWAIT
            # | WHEN |
            return_value = func(task, **firmware_update_args)
            # | THEN |
            attach_vmedia_mock.assert_any_call(
                task.node, 'CDROM', 'any_valid_url')
            self.assertEqual(exp_ret_state, return_value)
            execute_mock.assert_called_once_with(task, step, step_type)

    def test_write_firmware_sum_mode_without_component_clean(self):
        self._test_write_firmware_sum_mode_without_component(
            step_type='clean')

    def test_write_firmware_sum_mode_without_component_deploy(self):
        self._test_write_firmware_sum_mode_without_component(
            step_type='deploy')

    def _test_write_firmware_sum_mode_invalid_component(self,
                                                        step_type='clean'):
        # | GIVEN |
        firmware_update_args = {
            'url': 'any_valid_url',
            'checksum': 'xxxx',
            'components': ['CP02345']}
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            # | WHEN & THEN |
            if step_type == 'clean':
                func = task.driver.management.update_firmware_sum
            else:
                func = task.driver.management.flash_firmware_sum
            self.assertRaises(exception.InvalidParameterValue,
                              func, task, **firmware_update_args)

    def test_write_firmware_sum_mode_invalid_component_clean(self):
        self._test_write_firmware_sum_mode_invalid_component(
            step_type='clean')

    def test_write_firmware_sum_mode_invalid_component_deploy(self):
        self._test_write_firmware_sum_mode_invalid_component(
            step_type='deploy')

    @mock.patch.object(driver_utils, 'store_ramdisk_logs', autospec=True)
    def _test__write_firmware_sum_final_with_logs(self, store_mock,
                                                  step_type='clean'):
        self.config(deploy_logs_collect='always', group='agent')
        firmware_update_args = {
            'url': 'any_valid_url',
            'checksum': 'xxxx'}
        step = {'interface': 'management',
                'args': firmware_update_args}
        if step_type == 'clean':
            step['step'] = 'update_firmware_sum'
            node_state = states.CLEANWAIT
            command = {
                'command_status': 'SUCCEEDED',
                'command_result': {
                    'clean_result': {'Log Data': 'aaaabbbbcccdddd'},
                    'clean_step': step,
                }
            }
            exp_label = 'update_firmware_sum'
        else:
            step['step'] = 'flash_firmware_sum'
            node_state = states.DEPLOYWAIT
            command = {
                'command_status': 'SUCCEEDED',
                'command_result': {
                    'deploy_result': {'Log Data': 'aaaabbbbcccdddd'},
                    'deploy_step': step,
                }
            }
            exp_label = 'flash_firmware_sum'
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            task.node.provision_state = node_state
            task.driver.management._update_firmware_sum_final(
                task, command)
            store_mock.assert_called_once_with(task.node, 'aaaabbbbcccdddd',
                                               label=exp_label)

    def test__write_firmware_sum_final_with_logs_clean(self):
        self._test__write_firmware_sum_final_with_logs(step_type='clean')

    def test__write_firmware_sum_final_with_logs_deploy(self):
        self._test__write_firmware_sum_final_with_logs(step_type='deploy')

    @mock.patch.object(driver_utils, 'store_ramdisk_logs', autospec=True)
    def _test__write_firmware_sum_final_without_logs(self, store_mock,
                                                     step_type='clean'):
        self.config(deploy_logs_collect='on_failure', group='agent')
        firmware_update_args = {
            'url': 'any_valid_url',
            'checksum': 'xxxx'}
        step = {'interface': 'management',
                'args': firmware_update_args}
        if step_type == 'clean':
            step['step'] = 'update_firmware_sum'
            command = {
                'command_status': 'SUCCEEDED',
                'command_result': {
                    'clean_result': {'Log Data': 'aaaabbbbcccdddd'},
                    'clean_step': step,
                }
            }
        else:
            step['step'] = 'flash_firmware_sum'
            command = {
                'command_status': 'SUCCEEDED',
                'command_result': {
                    'deploy_result': {'Log Data': 'aaaabbbbcccdddd'},
                    'deploy_step': step,
                }
            }
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            task.driver.management._update_firmware_sum_final(
                task, command)
        self.assertFalse(store_mock.called)

    def test__write_firmware_sum_final_without_logs_clean(self):
        self._test__write_firmware_sum_final_without_logs(step_type='clean')

    def test__write_firmware_sum_final_without_logs_deploy(self):
        self._test__write_firmware_sum_final_without_logs(step_type='deploy')

    @mock.patch.object(ilo_management, 'LOG', spec_set=True, autospec=True)
    @mock.patch.object(driver_utils, 'store_ramdisk_logs', autospec=True)
    def _test__write_firmware_sum_final_swift_error(self, store_mock,
                                                    log_mock,
                                                    step_type='clean'):
        self.config(deploy_logs_collect='always', group='agent')
        firmware_update_args = {
            'url': 'any_valid_url',
            'checksum': 'xxxx'}
        step = {'interface': 'management',
                'args': firmware_update_args}
        if step_type == 'clean':
            step['step'] = 'update_firmware_sum'
            node_state = states.CLEANWAIT
            command = {
                'command_status': 'SUCCEEDED',
                'command_result': {
                    'clean_result': {'Log Data': 'aaaabbbbcccdddd'},
                    'clean_step': step,
                }
            }
        else:
            step['step'] = 'flash_firmware_sum'
            node_state = states.DEPLOYWAIT
            command = {
                'command_status': 'SUCCEEDED',
                'command_result': {
                    'deploy_result': {'Log Data': 'aaaabbbbcccdddd'},
                    'deploy_step': step,
                }
            }
        store_mock.side_effect = exception.SwiftOperationError('Error')

        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            task.node.provision_state = node_state
            task.driver.management._update_firmware_sum_final(
                task, command)
        self.assertTrue(log_mock.error.called)

    def test__write_firmware_sum_final_swift_error_clean(self):
        self._test__write_firmware_sum_final_swift_error(step_type='clean')

    def test__write_firmware_sum_final_swift_error_deploy(self):
        self._test__write_firmware_sum_final_swift_error(step_type='deploy')

    @mock.patch.object(ilo_management, 'LOG', spec_set=True, autospec=True)
    @mock.patch.object(driver_utils, 'store_ramdisk_logs', autospec=True)
    def _test__write_firmware_sum_final_environment_error(self, store_mock,
                                                          log_mock,
                                                          step_type='clean'):
        self.config(deploy_logs_collect='always', group='agent')
        firmware_update_args = {
            'url': 'any_valid_url',
            'checksum': 'xxxx'}
        step = {'interface': 'management',
                'args': firmware_update_args}
        if step_type == 'clean':
            step['step'] = 'update_firmware_sum'
            node_state = states.CLEANWAIT
            command = {
                'command_status': 'SUCCEEDED',
                'command_result': {
                    'clean_result': {'Log Data': 'aaaabbbbcccdddd'},
                    'clean_step': step,
                }
            }
        else:
            step['step'] = 'flash_firmware_sum'
            node_state = states.DEPLOYWAIT
            command = {
                'command_status': 'SUCCEEDED',
                'command_result': {
                    'deploy_result': {'Log Data': 'aaaabbbbcccdddd'},
                    'deploy_step': step,
                }
            }
        store_mock.side_effect = EnvironmentError('Error')

        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            task.node.provision_state = node_state
            task.driver.management._update_firmware_sum_final(
                task, command)
        self.assertTrue(log_mock.exception.called)

    def test__write_firmware_sum_final_environment_error_clean(self):
        self._test__write_firmware_sum_final_environment_error(
            step_type='clean')

    def test__write_firmware_sum_final_environment_error_deploy(self):
        self._test__write_firmware_sum_final_environment_error(
            step_type='deploy')

    @mock.patch.object(ilo_management, 'LOG', spec_set=True, autospec=True)
    @mock.patch.object(driver_utils, 'store_ramdisk_logs', autospec=True)
    def _test__write_firmware_sum_final_unknown_exception(self, store_mock,
                                                          log_mock,
                                                          step_type='clean'):
        self.config(deploy_logs_collect='always', group='agent')
        firmware_update_args = {
            'url': 'any_valid_url',
            'checksum': 'xxxx'}
        step = {'interface': 'management',
                'args': firmware_update_args}
        if step_type == 'clean':
            step['step'] = 'update_firmware_sum'
            node_state = states.CLEANWAIT
            command = {
                'command_status': 'SUCCEEDED',
                'command_result': {
                    'clean_result': {'Log Data': 'aaaabbbbcccdddd'},
                    'clean_step': step,
                }
            }
        else:
            step['step'] = 'flash_firmware_sum'
            node_state = states.DEPLOYWAIT
            command = {
                'command_status': 'SUCCEEDED',
                'command_result': {
                    'deploy_result': {'Log Data': 'aaaabbbbcccdddd'},
                    'deploy_step': step,
                }
            }
        store_mock.side_effect = Exception('Error')

        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            task.node.provision_state = node_state
            task.driver.management._update_firmware_sum_final(
                task, command)
        self.assertTrue(log_mock.exception.called)

    def test__write_firmware_sum_final_unknown_exception_clean(self):
        self._test__write_firmware_sum_final_unknown_exception(
            step_type='clean')

    def test__write_firmware_sum_final_unknown_exception_deploy(self):
        self._test__write_firmware_sum_final_unknown_exception(
            step_type='deploy')

    @mock.patch.object(ilo_common, 'get_ilo_object', spec_set=True,
                       autospec=True)
    def test_set_iscsi_boot_target_with_auth(self, get_ilo_object_mock):
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            vol_id = uuidutils.generate_uuid()
            obj_utils.create_test_volume_target(
                self.context, node_id=self.node.id, volume_type='iscsi',
                boot_index=0, volume_id='1234', uuid=vol_id,
                properties={'target_lun': 0,
                            'target_portal': 'fake_host:3260',
                            'target_iqn': 'fake_iqn',
                            'auth_username': 'fake_username',
                            'auth_password': 'fake_password'})
            driver_internal_info = task.node.driver_internal_info
            driver_internal_info['boot_from_volume'] = vol_id
            task.node.driver_internal_info = driver_internal_info
            task.node.save()
            ilo_object_mock = get_ilo_object_mock.return_value
            task.driver.management.set_iscsi_boot_target(task)
            ilo_object_mock.set_iscsi_info.assert_called_once_with(
                'fake_iqn', 0, 'fake_host', '3260',
                auth_method='CHAP', username='fake_username',
                password='fake_password',
                macs=['11:22:33:44:55:66', '11:22:33:44:55:67'])

    @mock.patch.object(ilo_common, 'get_ilo_object', spec_set=True,
                       autospec=True)
    def test_set_iscsi_boot_target_without_auth(self, get_ilo_object_mock):
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            vol_id = uuidutils.generate_uuid()
            obj_utils.create_test_volume_target(
                self.context, node_id=self.node.id, volume_type='iscsi',
                boot_index=0, volume_id='1234', uuid=vol_id,
                properties={'target_lun': 0,
                            'target_portal': 'fake_host:3260',
                            'target_iqn': 'fake_iqn'})
            driver_internal_info = task.node.driver_internal_info
            driver_internal_info['boot_from_volume'] = vol_id
            task.node.driver_internal_info = driver_internal_info
            task.node.save()
            ilo_object_mock = get_ilo_object_mock.return_value
            task.driver.management.set_iscsi_boot_target(task)
            ilo_object_mock.set_iscsi_info.assert_called_once_with(
                'fake_iqn', 0, 'fake_host', '3260', auth_method=None,
                password=None, username=None,
                macs=['11:22:33:44:55:66', '11:22:33:44:55:67'])

    @mock.patch.object(ilo_common, 'get_ilo_object', spec_set=True,
                       autospec=True)
    def test_set_iscsi_boot_target_failed(self, get_ilo_object_mock):
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            vol_id = uuidutils.generate_uuid()
            obj_utils.create_test_volume_target(
                self.context, node_id=self.node.id, volume_type='iscsi',
                boot_index=0, volume_id='1234', uuid=vol_id,
                properties={'target_lun': 0,
                            'target_portal': 'fake_host:3260',
                            'target_iqn': 'fake_iqn',
                            'auth_username': 'fake_username',
                            'auth_password': 'fake_password'})
            driver_internal_info = task.node.driver_internal_info
            driver_internal_info['boot_from_volume'] = vol_id
            task.node.driver_internal_info = driver_internal_info
            task.node.save()
            ilo_object_mock = get_ilo_object_mock.return_value
            ilo_object_mock.set_iscsi_info.side_effect = (
                ilo_error.IloError('error'))
            self.assertRaises(exception.IloOperationError,
                              task.driver.management.set_iscsi_boot_target,
                              task)

    def test_set_iscsi_boot_target_missed_properties(self):
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            vol_id = uuidutils.generate_uuid()
            obj_utils.create_test_volume_target(
                self.context, node_id=self.node.id, volume_type='iscsi',
                boot_index=0, volume_id='1234', uuid=vol_id,
                properties={'target_iqn': 'fake_iqn',
                            'auth_username': 'fake_username',
                            'auth_password': 'fake_password'})
            driver_internal_info = task.node.driver_internal_info
            driver_internal_info['boot_from_volume'] = vol_id
            task.node.driver_internal_info = driver_internal_info
            task.node.save()
            self.assertRaises(exception.MissingParameterValue,
                              task.driver.management.set_iscsi_boot_target,
                              task)

    @mock.patch.object(ilo_common, 'get_ilo_object', spec_set=True,
                       autospec=True)
    def test_set_iscsi_boot_target_in_bios(self, get_ilo_object_mock):
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            vol_id = uuidutils.generate_uuid()
            obj_utils.create_test_volume_target(
                self.context, node_id=self.node.id, volume_type='iscsi',
                boot_index=0, volume_id='1234', uuid=vol_id,
                properties={'target_lun': 0,
                            'target_portal': 'fake_host:3260',
                            'target_iqn': 'fake_iqn',
                            'auth_username': 'fake_username',
                            'auth_password': 'fake_password'})
            driver_internal_info = task.node.driver_internal_info
            driver_internal_info['boot_from_volume'] = vol_id
            task.node.driver_internal_info = driver_internal_info
            task.node.save()
            ilo_object_mock = get_ilo_object_mock.return_value
            ilo_object_mock.set_iscsi_info.side_effect = (
                ilo_error.IloCommandNotSupportedInBiosError('error'))
            self.assertRaises(exception.IloOperationNotSupported,
                              task.driver.management.set_iscsi_boot_target,
                              task)

    @mock.patch.object(ilo_common, 'get_ilo_object', spec_set=True,
                       autospec=True)
    def test_clear_iscsi_boot_target(self, get_ilo_object_mock):
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            ilo_object_mock = get_ilo_object_mock.return_value

            task.driver.management.clear_iscsi_boot_target(task)
            ilo_object_mock.unset_iscsi_info.assert_called_once_with(
                macs=['11:22:33:44:55:66', '11:22:33:44:55:67'])

    @mock.patch.object(ilo_common, 'get_ilo_object', spec_set=True,
                       autospec=True)
    def test_clear_iscsi_boot_target_failed(self, get_ilo_object_mock):
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            ilo_object_mock = get_ilo_object_mock.return_value
            ilo_object_mock.unset_iscsi_info.side_effect = (
                ilo_error.IloError('error'))
            self.assertRaises(exception.IloOperationError,
                              task.driver.management.clear_iscsi_boot_target,
                              task)

    @mock.patch.object(ilo_common, 'get_ilo_object', spec_set=True,
                       autospec=True)
    def test_clear_iscsi_boot_target_in_bios(self, get_ilo_object_mock):
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            ilo_object_mock = get_ilo_object_mock.return_value
            ilo_object_mock.unset_iscsi_info.side_effect = (
                ilo_error.IloCommandNotSupportedInBiosError('error'))
            self.assertRaises(exception.IloOperationNotSupported,
                              task.driver.management.clear_iscsi_boot_target,
                              task)

    @mock.patch.object(ilo_common, 'get_ilo_object', spec_set=True,
                       autospec=True)
    def test_inject_nmi(self, get_ilo_object_mock):
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            ilo_object_mock = get_ilo_object_mock.return_value

            task.driver.management.inject_nmi(task)
            ilo_object_mock.inject_nmi.assert_called_once()

    @mock.patch.object(ilo_common, 'get_ilo_object', spec_set=True,
                       autospec=True)
    def test_inject_nmi_failed(self, get_ilo_object_mock):
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            ilo_object_mock = get_ilo_object_mock.return_value
            ilo_object_mock.inject_nmi.side_effect = (
                ilo_error.IloError('error'))
            self.assertRaises(exception.IloOperationError,
                              task.driver.management.inject_nmi,
                              task)

    @mock.patch.object(ilo_common, 'get_ilo_object', spec_set=True,
                       autospec=True)
    def test_inject_nmi_not_supported(self, get_ilo_object_mock):
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            ilo_object_mock = get_ilo_object_mock.return_value
            ilo_object_mock.inject_nmi.side_effect = (
                ilo_error.IloCommandNotSupportedError('error'))
            self.assertRaises(exception.IloOperationNotSupported,
                              task.driver.management.inject_nmi,
                              task)

    @mock.patch.object(ilo_common, 'get_ilo_object', spec_set=True,
                       autospec=True)
    @ddt.data((ilo_common.SUPPORTED_BOOT_MODE_LEGACY_BIOS_ONLY,
               ['bios']),
              (ilo_common.SUPPORTED_BOOT_MODE_UEFI_ONLY,
               ['uefi']),
              (ilo_common.SUPPORTED_BOOT_MODE_LEGACY_BIOS_AND_UEFI,
               ['uefi', 'bios']))
    @ddt.unpack
    def test_get_supported_boot_modes(self, boot_modes_val,
                                      exp_boot_modes,
                                      get_ilo_object_mock):
        ilo_object_mock = get_ilo_object_mock.return_value
        ilo_object_mock.get_supported_boot_mode.return_value = boot_modes_val
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            supported_boot_modes = (
                task.driver.management.get_supported_boot_modes(task))
            self.assertEqual(exp_boot_modes, supported_boot_modes)

    @mock.patch.object(ilo_common, 'set_boot_mode', spec_set=True,
                       autospec=True)
    @mock.patch.object(ilo_management.IloManagement,
                       'get_supported_boot_modes',
                       spec_set=True, autospec=True)
    def test_set_boot_mode(self, supp_boot_modes_mock,
                           set_boot_mode_mock):
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            exp_boot_modes = [boot_modes.UEFI, boot_modes.LEGACY_BIOS]
            supp_boot_modes_mock.return_value = exp_boot_modes

            for mode in exp_boot_modes:
                task.driver.management.set_boot_mode(task, mode=mode)
                supp_boot_modes_mock.assert_called_once_with(mock.ANY, task)
                set_boot_mode_mock.assert_called_once_with(task.node, mode)
                set_boot_mode_mock.reset_mock()
                supp_boot_modes_mock.reset_mock()

    @mock.patch.object(ilo_common, 'get_ilo_object', spec_set=True,
                       autospec=True)
    @mock.patch.object(ilo_management.IloManagement,
                       'get_supported_boot_modes',
                       spec_set=True, autospec=True)
    def test_set_boot_mode_fail(self, supp_boot_modes_mock,
                                get_ilo_object_mock):
        ilo_mock_obj = get_ilo_object_mock.return_value
        ilo_mock_obj.get_pending_boot_mode.return_value = 'legacy'
        exc = ilo_error.IloError('error')
        ilo_mock_obj.set_pending_boot_mode.side_effect = exc
        exp_boot_modes = [boot_modes.UEFI, boot_modes.LEGACY_BIOS]
        supp_boot_modes_mock.return_value = exp_boot_modes
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            self.assertRaisesRegex(
                exception.IloOperationError, 'uefi as boot mode failed',
                task.driver.management.set_boot_mode, task, boot_modes.UEFI)
            supp_boot_modes_mock.assert_called_once_with(mock.ANY, task)

    @mock.patch.object(ilo_common, 'get_ilo_object', spec_set=True,
                       autospec=True)
    def test_get_boot_mode(self, get_ilo_object_mock):
        expected = 'bios'
        ilo_mock_obj = get_ilo_object_mock.return_value
        ilo_mock_obj.get_current_boot_mode.return_value = 'LEGACY'
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            response = task.driver.management.get_boot_mode(task)
            self.assertEqual(expected, response)

    @mock.patch.object(ilo_common, 'get_ilo_object', spec_set=True,
                       autospec=True)
    def test_get_boot_mode_fail(self, get_ilo_object_mock):
        ilo_mock_obj = get_ilo_object_mock.return_value
        exc = ilo_error.IloError('error')
        ilo_mock_obj.get_current_boot_mode.side_effect = exc
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            self.assertRaisesRegex(
                exception.IloOperationError, 'Get current boot mode',
                task.driver.management.get_boot_mode, task)


class Ilo5ManagementTestCase(db_base.DbTestCase):

    def setUp(self):
        super(Ilo5ManagementTestCase, self).setUp()
        self.driver = mock.Mock(management=ilo_management.Ilo5Management())

        n = {
            'driver': 'ilo5',
            'driver_info': INFO_DICT,
        }
        self.config(enabled_hardware_types=['ilo5'],
                    enabled_boot_interfaces=['ilo-virtual-media'],
                    enabled_console_interfaces=['ilo'],
                    enabled_deploy_interfaces=['direct'],
                    enabled_inspect_interfaces=['ilo'],
                    enabled_management_interfaces=['ilo5'],
                    enabled_power_interfaces=['ilo'],
                    enabled_raid_interfaces=['ilo5'])
        self.node = obj_utils.create_test_node(self.context, **n)

    @mock.patch.object(deploy_utils, 'build_agent_options',
                       autospec=True)
    @mock.patch.object(ilo_common, 'get_ilo_object', autospec=True)
    @mock.patch.object(manager_utils, 'node_power_action', autospec=True)
    def test_erase_devices_hdd(self, mock_power, ilo_mock, build_agent_mock):
        self.node.clean_step = {'interface': 'management',
                                'step': 'erase_devices'}
        self.node.save()
        ilo_mock_object = ilo_mock.return_value
        ilo_mock_object.get_available_disk_types.return_value = ['HDD']
        build_agent_mock.return_value = []
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            result = task.driver.management.erase_devices(task)
            self.assertTrue(
                task.node.driver_internal_info.get(
                    'ilo_disk_erase_hdd_check'))
            self.assertTrue(
                task.node.driver_internal_info.get(
                    'cleaning_reboot'))
            self.assertFalse(
                task.node.driver_internal_info.get(
                    'skip_current_clean_step'))
            ilo_mock_object.do_disk_erase.assert_called_once_with(
                'HDD', 'overwrite')
            self.assertEqual(states.CLEANWAIT, result)
            mock_power.assert_called_once_with(task, states.REBOOT, None)

    @mock.patch.object(deploy_utils, 'build_agent_options',
                       autospec=True)
    @mock.patch.object(ilo_common, 'get_ilo_object', autospec=True)
    @mock.patch.object(manager_utils, 'node_power_action', autospec=True)
    def test_erase_devices_ssd(self, mock_power, ilo_mock, build_agent_mock):
        self.node.clean_step = {'interface': 'management',
                                'step': 'erase_devices'}
        self.node.save()
        ilo_mock_object = ilo_mock.return_value
        ilo_mock_object.get_available_disk_types.return_value = ['SSD']
        build_agent_mock.return_value = []
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            result = task.driver.management.erase_devices(task)
            self.assertTrue(
                task.node.driver_internal_info.get(
                    'ilo_disk_erase_ssd_check'))
            self.assertTrue(
                task.node.driver_internal_info.get(
                    'ilo_disk_erase_hdd_check'))
            self.assertTrue(
                task.node.driver_internal_info.get(
                    'cleaning_reboot'))
            self.assertFalse(
                task.node.driver_internal_info.get(
                    'skip_current_clean_step'))
            ilo_mock_object.do_disk_erase.assert_called_once_with(
                'SSD', 'block')
            self.assertEqual(states.CLEANWAIT, result)
            mock_power.assert_called_once_with(task, states.REBOOT, None)

    @mock.patch.object(deploy_utils, 'build_agent_options',
                       autospec=True)
    @mock.patch.object(ilo_common, 'get_ilo_object', autospec=True)
    @mock.patch.object(manager_utils, 'node_power_action', autospec=True)
    def test_erase_devices_ssd_when_hdd_done(self, mock_power, ilo_mock,
                                             build_agent_mock):
        self.node.clean_step = {'interface': 'management',
                                'step': 'erase_devices'}
        self.node.save()
        build_agent_mock.return_value = []
        ilo_mock_object = ilo_mock.return_value
        ilo_mock_object.get_available_disk_types.return_value = ['HDD', 'SSD']
        self.node.driver_internal_info = {'ilo_disk_erase_hdd_check': True}
        self.node.save()
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            result = task.driver.management.erase_devices(task)
            self.assertTrue(
                task.node.driver_internal_info.get(
                    'ilo_disk_erase_hdd_check'))
            self.assertTrue(
                task.node.driver_internal_info.get(
                    'ilo_disk_erase_ssd_check'))
            self.assertTrue(
                task.node.driver_internal_info.get(
                    'cleaning_reboot'))
            self.assertFalse(
                task.node.driver_internal_info.get(
                    'skip_current_clean_step'))
            ilo_mock_object.do_disk_erase.assert_called_once_with(
                'SSD', 'block')
            self.assertEqual(states.CLEANWAIT, result)
            mock_power.assert_called_once_with(task, states.REBOOT, None)

    @mock.patch.object(ilo_management.LOG, 'info', autospec=True)
    @mock.patch.object(ilo_management.Ilo5Management,
                       '_wait_for_disk_erase_status', autospec=True)
    @mock.patch.object(ilo_common, 'get_ilo_object', autospec=True)
    def test_erase_devices_completed(self, ilo_mock, disk_status_mock,
                                     log_mock):
        self.node.clean_step = {'interface': 'management',
                                'step': 'erase_devices'}
        self.node.save()
        ilo_mock_object = ilo_mock.return_value
        ilo_mock_object.get_available_disk_types.return_value = ['HDD', 'SSD']
        disk_status_mock.return_value = True
        self.node.driver_internal_info = {'ilo_disk_erase_hdd_check': True,
                                          'ilo_disk_erase_ssd_check': True}
        self.node.save()
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            task.driver.management.erase_devices(task)
            self.assertFalse(
                task.node.driver_internal_info.get(
                    'ilo_disk_erase_hdd_check'))
            self.assertFalse(
                task.node.driver_internal_info.get(
                    'ilo_disk_erase_hdd_check'))
            self.assertTrue(log_mock.called)

    @mock.patch.object(deploy_utils, 'build_agent_options',
                       autospec=True)
    @mock.patch.object(ilo_common, 'get_ilo_object', autospec=True)
    @mock.patch.object(manager_utils, 'node_power_action', autospec=True)
    def test_erase_devices_hdd_with_erase_pattern_zero(
            self, mock_power, ilo_mock, build_agent_mock):
        self.node.clean_step = {'interface': 'management',
                                'step': 'erase_devices'}
        self.node.save()
        ilo_mock_object = ilo_mock.return_value
        ilo_mock_object.get_available_disk_types.return_value = ['HDD']
        build_agent_mock.return_value = []
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            result = task.driver.management.erase_devices(
                task, erase_pattern={'hdd': 'zero', 'ssd': 'zero'})
            self.assertTrue(
                task.node.driver_internal_info.get(
                    'ilo_disk_erase_hdd_check'))
            self.assertTrue(
                task.node.driver_internal_info.get(
                    'cleaning_reboot'))
            self.assertFalse(
                task.node.driver_internal_info.get(
                    'skip_current_clean_step'))
            ilo_mock_object.do_disk_erase.assert_called_once_with(
                'HDD', 'zero')
            self.assertEqual(states.CLEANWAIT, result)
            mock_power.assert_called_once_with(task, states.REBOOT, None)

    @mock.patch.object(ilo_management.LOG, 'info', autospec=True)
    @mock.patch.object(ilo_common, 'get_ilo_object', autospec=True)
    def test_erase_devices_when_no_drive_available(
            self, ilo_mock, log_mock):
        self.node.clean_step = {'interface': 'management',
                                'step': 'erase_devices'}
        self.node.save()
        ilo_mock_object = ilo_mock.return_value
        ilo_mock_object.get_available_disk_types.return_value = []
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            task.driver.management.erase_devices(task)
            self.assertTrue(log_mock.called)

    def test_erase_devices_hdd_with_invalid_format_erase_pattern(
            self):
        self.node.clean_step = {'interface': 'management',
                                'step': 'erase_devices'}
        self.node.save()
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            self.assertRaises(exception.InvalidParameterValue,
                              task.driver.management.erase_devices,
                              task, erase_pattern=123)

    def test_erase_devices_hdd_with_invalid_device_type_erase_pattern(
            self):
        self.node.clean_step = {'interface': 'management',
                                'step': 'erase_devices'}
        self.node.save()
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            self.assertRaises(exception.InvalidParameterValue,
                              task.driver.management.erase_devices,
                              task, erase_pattern={'xyz': 'block'})

    def test_erase_devices_hdd_with_invalid_erase_pattern(
            self):
        self.node.clean_step = {'interface': 'management',
                                'step': 'erase_devices'}
        self.node.save()
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            self.assertRaises(exception.InvalidParameterValue,
                              task.driver.management.erase_devices,
                              task, erase_pattern={'ssd': 'xyz'})

    @mock.patch.object(ilo_common, 'get_ilo_object', autospec=True)
    @mock.patch.object(manager_utils, 'cleaning_error_handler',
                       autospec=True)
    def test_erase_devices_hdd_ilo_error(self, clean_err_handler_mock,
                                         ilo_mock):
        self.node.clean_step = {'interface': 'management',
                                'step': 'erase_devices'}
        self.node.save()
        ilo_mock_object = ilo_mock.return_value
        ilo_mock_object.get_available_disk_types.return_value = ['HDD']
        exc = ilo_error.IloError('error')
        ilo_mock_object.do_disk_erase.side_effect = exc
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            task.driver.management.erase_devices(task)
            ilo_mock_object.do_disk_erase.assert_called_once_with(
                'HDD', 'overwrite')
            self.assertNotIn('ilo_disk_erase_hdd_check',
                             task.node.driver_internal_info)
            self.assertNotIn('ilo_disk_erase_ssd_check',
                             task.node.driver_internal_info)
            self.assertNotIn('cleaning_reboot',
                             task.node.driver_internal_info)
            self.assertNotIn('skip_current_clean_step',
                             task.node.driver_internal_info)
            clean_err_handler_mock.assert_called_once_with(
                task,
                ("Out-of-band sanitize disk erase job failed for node %s. "
                 "Message: 'error'." % task.node.uuid),
                errmsg=exc)

    @mock.patch.object(manager_utils, 'node_power_action', autospec=True)
    @mock.patch.object(ilo_common, 'get_ilo_object', autospec=True)
    def test_one_button_secure_erase(self, ilo_mock, mock_power):
        ilo_mock_object = ilo_mock.return_value
        self.node.clean_step = {'step': 'one_button_secure_erase',
                                'interface': 'management'}
        self.node.save()
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            result = task.driver.management.one_button_secure_erase(task)
            self.assertTrue(
                ilo_mock_object.do_one_button_secure_erase.called)
            self.assertEqual(states.CLEANWAIT, result)
            mock_power.assert_called_once_with(task, states.REBOOT)
            self.assertEqual(task.node.maintenance, True)

    @mock.patch.object(ilo_common, 'get_ilo_object', autospec=True)
    @mock.patch.object(manager_utils, 'cleaning_error_handler',
                       autospec=True)
    def test_one_button_secure_erase_ilo_error(
            self, clean_err_handler_mock, ilo_mock):
        ilo_mock_object = ilo_mock.return_value
        self.node.clean_step = {'step': 'one_button_secure_erase',
                                'interface': 'management'}
        self.node.save()
        exc = ilo_error.IloError('error')
        ilo_mock_object.do_one_button_secure_erase.side_effect = exc
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            task.driver.management.one_button_secure_erase(task)
            clean_err_handler_mock.assert_called_once_with(
                task,
                ("One button secure erase job failed for node %s. "
                 "Message: 'error'." % task.node.uuid),
                errmsg=exc)
            self.assertTrue(
                ilo_mock_object.do_one_button_secure_erase.called)

    @mock.patch.object(deploy_utils, 'get_async_step_return_state',
                       spec_set=True, autospec=True)
    @mock.patch.object(deploy_utils, 'set_async_step_flags',
                       spec_set=True, autospec=True)
    @mock.patch.object(deploy_utils, 'build_agent_options',
                       spec_set=True, autospec=True)
    @mock.patch.object(manager_utils, 'node_power_action', autospec=True)
    @mock.patch.object(ilo_boot.IloVirtualMediaBoot, 'prepare_ramdisk',
                       spec_set=True, autospec=True)
    @mock.patch.object(ilo_common, 'clear_certificates', spec_set=True,
                       autospec=True)
    def test_clear_ca_certificates(self, clear_certs_mock,
                                   prepare_ramdisk_mock, node_power_mock,
                                   build_agent_mock, set_async_mock,
                                   get_async_ret_mock):
        self.node.clean_step = {'interface': 'management',
                                'step': 'clear_ca_certificates'}
        self.node.save()
        build_agent_mock.return_value = {'a': 'x', 'b': 'y'}
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:

            certificate_files = ["/path/to/certsA", "/path/to/certsB"]

            task.driver.management.clear_ca_certificates(
                task, certificate_files)
            clear_certs_mock.assert_called_once_with(
                task, certificate_files)
            self.assertTrue(task.node.driver_internal_info.get(
                'clear_ca_certs_flag'))
            build_agent_mock.assert_called_once_with(task.node)
            prepare_ramdisk_mock.assert_called_once_with(
                mock.ANY, task, {'a': 'x', 'b': 'y'})
            node_power_mock.assert_called_once_with(task, states.REBOOT)
            set_async_mock.assert_called_once_with(
                task.node, reboot=True, skip_current_step=False)
            get_async_ret_mock.assert_called_once_with(task.node)

    @mock.patch.object(deploy_utils, 'get_async_step_return_state',
                       spec_set=True, autospec=True)
    @mock.patch.object(deploy_utils, 'set_async_step_flags',
                       spec_set=True, autospec=True)
    @mock.patch.object(deploy_utils, 'build_agent_options',
                       spec_set=True, autospec=True)
    @mock.patch.object(manager_utils, 'node_power_action', autospec=True)
    @mock.patch.object(ilo_boot.IloVirtualMediaBoot, 'prepare_ramdisk',
                       spec_set=True, autospec=True)
    @mock.patch.object(ilo_common, 'clear_certificates', spec_set=True,
                       autospec=True)
    def test_clear_ca_certificates_clear_flag(
            self, clear_certs_mock, prepare_ramdisk_mock, node_power_mock,
            build_agent_mock, set_async_mock, get_async_ret_mock):
        self.node.clean_step = {'interface': 'management',
                                'step': 'clear_ca_certificates'}

        self.node.save()
        build_agent_mock.return_value = {'a': 'x', 'b': 'y'}
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:

            certificate_files = ["/path/to/certsA", "/path/to/certsB"]
            driver_internal_info = task.node.driver_internal_info
            driver_internal_info['clear_ca_certs_flag'] = True
            task.node.driver_internal_info = driver_internal_info

            task.driver.management.clear_ca_certificates(
                task, certificate_files)
            clear_certs_mock.assert_not_called()

            build_agent_mock.assert_not_called()
            prepare_ramdisk_mock.assert_not_called()
            node_power_mock.assert_not_called()
            set_async_mock.assert_not_called()
            get_async_ret_mock.assert_not_called()

    @mock.patch.object(deploy_utils, 'get_async_step_return_state',
                       spec_set=True, autospec=True)
    @mock.patch.object(deploy_utils, 'set_async_step_flags',
                       spec_set=True, autospec=True)
    @mock.patch.object(deploy_utils, 'build_agent_options',
                       spec_set=True, autospec=True)
    @mock.patch.object(manager_utils, 'node_power_action', autospec=True)
    @mock.patch.object(ilo_boot.IloVirtualMediaBoot, 'prepare_ramdisk',
                       spec_set=True, autospec=True)
    @mock.patch.object(ilo_common, 'clear_certificates', spec_set=True,
                       autospec=True)
    def test_clear_ca_certificates_ilo_operation_not_supported(
            self, clear_certs_mock, prepare_ramdisk_mock, node_power_mock,
            build_agent_mock, set_async_mock, get_async_ret_mock):
        self.node.clean_step = {'interface': 'management',
                                'step': 'clear_ca_certificates'}
        self.node.save()
        exc = exception.IloOperationNotSupported('error')
        clear_certs_mock.side_effect = exc

        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:

            certificate_files = ["/path/to/certsA", "/path/to/certsB"]

            self.assertRaises(exception.NodeCleaningFailure,
                              task.driver.management.clear_ca_certificates,
                              task, certificate_files)
            build_agent_mock.assert_not_called()
            prepare_ramdisk_mock.assert_not_called()
            node_power_mock.assert_not_called()
            set_async_mock.assert_not_called()
            get_async_ret_mock.assert_not_called()

    @mock.patch.object(ilo_common, 'clear_certificates', spec_set=True,
                       autospec=True)
    def test_clear_ca_certificates_ilo_operation_error(self, clear_certs_mock):
        self.node.deploy_step = {'interface': 'management',
                                 'step': 'clear_ca_certificates'}
        self.node.save()

        exc = exception.IloOperationError('error')
        clear_certs_mock.side_effect = exc

        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:

            certificate_files = ["/path/to/certsA", "/path/to/certsB"]

            self.assertRaises(exception.InstanceDeployFailure,
                              task.driver.management.clear_ca_certificates,
                              task, certificate_files)



# Copyright 2010 United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All Rights Reserved.
# Copyright (c) 2010 Citrix Systems, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Handling of VM disk images.
"""

import os
import shutil
import time

from oslo_concurrency import processutils
from oslo_log import log as logging
from oslo_utils import fileutils
from oslo_utils.imageutils import format_inspector as image_format_inspector
import pycdlib

from ironic.common import checksum_utils
from ironic.common import exception
from ironic.common.glance_service import service_utils as glance_utils
from ironic.common.i18n import _
from ironic.common import image_service as service
from ironic.common import qemu_img
from ironic.common import utils
from ironic.conf import CONF

LOG = logging.getLogger(__name__)


def _create_root_fs(root_directory, files_info):
    """Creates a filesystem root in given directory.

    Given a mapping of absolute path of files to their relative paths
    within the filesystem, this method copies the files to their
    destination.

    :param root_directory: the filesystem root directory.
    :param files_info: A dict containing absolute path of file to be copied
                       or its content as bytes -> relative path within
                       the vfat image. For example::
                        {
                        '/absolute/path/to/file': 'relative/path/within/root',
                        b'{"some": "json"}': 'another/relative/path'
                        ...
                        }
    :raises: OSError, if creation of any directory failed.
    :raises: IOError, if copying any of the files failed.
    """
    for src_file, path in files_info.items():
        LOG.debug('Injecting %(path)s into an ISO from %(source)r',
                  {'path': path, 'source': src_file})
        target_file = os.path.join(root_directory, path)
        dirname = os.path.dirname(target_file)
        if dirname:
            os.makedirs(dirname, exist_ok=True)

        if isinstance(src_file, bytes):
            with open(target_file, 'wb') as fp:
                fp.write(src_file)
        else:
            shutil.copyfile(src_file, target_file)


def create_vfat_image(output_file, files_info=None, parameters=None,
                      parameters_file='parameters.txt', fs_size_kib=100):
    """Creates the fat fs image on the desired file.

    This method copies the given files to a root directory (optional),
    writes the parameters specified to the parameters file within the
    root directory (optional), and then creates a vfat image of the root
    directory.

    :param output_file: The path to the file where the fat fs image needs
                        to be created.
    :param files_info: A dict containing absolute path of file to be copied
                       -> relative path within the vfat image. For example::

                        {
                        '/absolute/path/to/file' -> 'relative/path/within/root'
                        ...
                        }
    :param parameters: A dict containing key-value pairs of parameters.
    :param parameters_file: The filename for the parameters file.
    :param fs_size_kib: size of the vfat filesystem in KiB.
    :raises: ImageCreationFailed, if image creation failed while doing any
             of filesystem manipulation activities like creating dirs,
             mounting, creating filesystem, copying files, etc.
    """
    try:
        # TODO(sbaker): use utils.dd when rootwrap has been removed
        utils.execute('dd', 'if=/dev/zero', 'of=%s' % output_file, 'count=1',
                      'bs=%dKiB' % fs_size_kib)
    except processutils.ProcessExecutionError as e:
        raise exception.ImageCreationFailed(image_type='vfat', error=e)

    with utils.tempdir() as tmpdir:

        try:
            # The label helps ramdisks to find the partition containing
            # the parameters (by using /dev/disk/by-label/ir-vfd-dev).
            # NOTE: FAT filesystem label can be up to 11 characters long.
            # TODO(sbaker): use utils.mkfs when rootwrap has been removed
            utils.execute('mkfs', '-t', 'vfat', '-n',
                          'ir-vfd-dev', output_file)
        except processutils.ProcessExecutionError as e:
            raise exception.ImageCreationFailed(image_type='vfat', error=e)

        try:
            if files_info:
                _create_root_fs(tmpdir, files_info)

            if parameters:
                parameters_file = os.path.join(tmpdir, parameters_file)
                params_list = ['%(key)s=%(val)s' % {'key': k, 'val': v}
                               for k, v in parameters.items()]
                file_contents = '\n'.join(params_list)
                utils.write_to_file(parameters_file, file_contents)

            file_list = os.listdir(tmpdir)

            if not file_list:
                return

            file_list = [os.path.join(tmpdir, item) for item in file_list]

            # use mtools to copy the files into the image in a single
            # operation
            utils.execute('mcopy', '-s', *file_list, '-i', output_file, '::')

        except Exception as e:
            LOG.exception("vfat image creation failed. Error: %s", e)
            raise exception.ImageCreationFailed(image_type='vfat', error=e)


def _generate_cfg(kernel_params, template, options):
    """Generates a isolinux or grub configuration file.

    Given a given a list of strings containing kernel parameters, this method
    returns the kernel cmdline string.
    :param kernel_params: a list of strings(each element being a string like
        'K=V' or 'K' or combination of them like 'K1=V1 K2 K3=V3') to be added
        as the kernel cmdline.
    :param template: the path of the config template file.
    :param options: a dictionary of keywords which need to be replaced in
                    template file to generate a proper config file.
    :returns: a string containing the contents of the isolinux configuration
        file.
    """
    options.update({'kernel_params': ' '.join(kernel_params or [])})
    return utils.render_template(template, options)


def _label(files_info):
    """Get a suitable label for the files.

    Returns "config-2" if the openstack metadata is present.
    """
    if any(x.startswith('openstack/') for x in files_info.values()):
        return 'config-2'
    else:
        return 'VMEDIA_BOOT_ISO'


def create_isolinux_image_for_bios(
        output_file, kernel, ramdisk, kernel_params=None, inject_files=None,
        publisher_id=None):
    """Creates an isolinux image on the specified file.

    Copies the provided kernel, ramdisk to a directory, generates the isolinux
    configuration file using the kernel parameters provided, and then generates
    a bootable ISO image.

    :param output_file: the path to the file where the iso image needs to be
        created.
    :param kernel: the kernel to use.
    :param ramdisk: the ramdisk to use.
    :param kernel_params: a list of strings(each element being a string like
        'K=V' or 'K' or combination of them like 'K1=V1,K2,...') to be added
        as the kernel cmdline.
    :param inject_files: Mapping of local source file paths to their location
        on the final ISO image.
    :param publisher_id: A value to set as the publisher identifier string
        in the ISO image to be generated.
    :raises: ImageCreationFailed, if image creation failed while copying files
        or while running command to generate iso.
    """
    ISOLINUX_BIN = 'isolinux/isolinux.bin'
    ISOLINUX_CFG = 'isolinux/isolinux.cfg'
    LDLINUX_SRC_DIRS = ['/usr/lib/syslinux/modules/bios',
                        '/usr/share/syslinux']
    LDLINUX_BIN = 'isolinux/ldlinux.c32'

    options = {'kernel': '/vmlinuz', 'ramdisk': '/initrd'}

    with utils.tempdir() as tmpdir:
        files_info = {
            kernel: 'vmlinuz',
            ramdisk: 'initrd',
            CONF.isolinux_bin: ISOLINUX_BIN,
        }
        if inject_files:
            files_info.update(inject_files)

        # ldlinux.c32 is required for syslinux 5.0 or later.
        if CONF.ldlinux_c32:
            ldlinux_src = CONF.ldlinux_c32
        else:
            for directory in LDLINUX_SRC_DIRS:
                ldlinux_src = os.path.join(directory, 'ldlinux.c32')
                if os.path.isfile(ldlinux_src):
                    break
            else:
                ldlinux_src = None
        if ldlinux_src:
            files_info[ldlinux_src] = LDLINUX_BIN

        try:
            _create_root_fs(tmpdir, files_info)

        except EnvironmentError as e:
            LOG.exception("Creating the filesystem root failed.")
            raise exception.ImageCreationFailed(image_type='iso', error=e)

        cfg = _generate_cfg(kernel_params,
                            CONF.isolinux_config_template, options)

        isolinux_cfg = os.path.join(tmpdir, ISOLINUX_CFG)
        utils.write_to_file(isolinux_cfg, cfg)

        # Set a publisher ID value to a string.
        pub_id = str(publisher_id)

        try:
            utils.execute('mkisofs', '-r', '-V', _label(files_info),
                          '-J', '-l', '-publisher', pub_id, '-no-emul-boot',
                          '-boot-load-size', '4', '-boot-info-table',
                          '-b', ISOLINUX_BIN, '-o', output_file, tmpdir)
        except processutils.ProcessExecutionError as e:
            LOG.exception("Creating ISO image failed.")
            raise exception.ImageCreationFailed(image_type='iso', error=e)


def create_esp_image_for_uefi(
        output_file, kernel, ramdisk, deploy_iso=None, esp_image=None,
        kernel_params=None, inject_files=None, publisher_id=None):
    """Creates an ESP image on the specified file.

    Copies the provided kernel, ramdisk and EFI system partition image (ESP) to
    a directory, generates the grub configuration file using kernel parameters
    and then generates a bootable ISO image for UEFI.

    :param output_file: the path to the file where the iso image needs to be
        created.
    :param kernel: the kernel to use.
    :param ramdisk: the ramdisk to use.
    :param deploy_iso: deploy ISO image to extract EFI system partition image
        from. If not specified, the `esp_image` option is required.
    :param esp_image: FAT12/16/32-formatted EFI system partition image
        containing the EFI boot loader (e.g. GRUB2) for each hardware
        architecture to boot. This image will be embedded into the ISO image.
        If not specified, the `deploy_iso` option is required.
    :param kernel_params: a list of strings(each element being a string like
        'K=V' or 'K' or combination of them like 'K1=V1,K2,...') to be added
        as the kernel cmdline.
    :param inject_files: Mapping of local source file paths to their location
        on the final ISO image.
    :param publisher_id: A value to set as the publisher identifier string
        in the ISO image to be generated.
    :raises: ImageCreationFailed, if image creation failed while copying files
        or while running command to generate iso.
    """
    EFIBOOT_LOCATION = 'boot/grub/efiboot.img'

    grub_options = {'linux': '/vmlinuz', 'initrd': '/initrd'}

    with utils.tempdir() as tmpdir:
        files_info = {
            kernel: 'vmlinuz',
            ramdisk: 'initrd',
        }
        if inject_files:
            files_info.update(inject_files)

        with utils.tempdir() as mountdir:
            # Open the deploy iso used to initiate deploy and copy the
            # efiboot.img i.e. boot loader to the current temporary
            # directory.
            if deploy_iso and not esp_image:
                uefi_path_info, e_img_rel_path, grub_rel_path = (
                    _get_deploy_iso_files(deploy_iso, mountdir))

                grub_cfg = os.path.join(tmpdir, grub_rel_path)

            # Use ELF boot loader provided
            elif esp_image and not deploy_iso:
                e_img_rel_path = EFIBOOT_LOCATION
                grub_rel_path = CONF.grub_config_path.lstrip(' ' + os.sep)
                grub_cfg = os.path.join(tmpdir, grub_rel_path)

                # Create an empty grub config file by copying /dev/null.
                # This is to avoid write failures when actual write of
                # config file happens. Write failures are caused if grub
                # config path does not exist on root file system.
                uefi_path_info = {
                    esp_image: e_img_rel_path,
                    '/dev/null': grub_rel_path
                }

            else:
                msg = _('Neither deploy ISO nor ESP image configured or '
                        'both of them configured')
                raise exception.ImageCreationFailed(
                    image_type='iso', error=msg)

            files_info.update(uefi_path_info)

            try:
                _create_root_fs(tmpdir, files_info)

            except EnvironmentError as e:
                LOG.exception("Creating the filesystem root failed.")
                raise exception.ImageCreationFailed(
                    image_type='iso', error=e)

            finally:
                if deploy_iso:
                    shutil.rmtree(mountdir)

        # Generate and copy grub config file.
        grub_conf = _generate_cfg(kernel_params,
                                  CONF.grub_config_template, grub_options)
        utils.write_to_file(grub_cfg, grub_conf)

        # Create the boot_iso.
        if publisher_id:
            args = ('mkisofs', '-r', '-V', _label(files_info),
                    '-l', '-publisher', publisher_id, '-e', e_img_rel_path,
                    '-no-emul-boot', '-o', output_file,
                    tmpdir)
        else:
            args = ('mkisofs', '-r', '-V', _label(files_info),
                    '-l', '-e', e_img_rel_path,
                    '-no-emul-boot', '-o', output_file,
                    tmpdir)
        try:
            utils.execute(*args)

        except processutils.ProcessExecutionError as e:
            LOG.exception("Creating ISO image failed.")
            raise exception.ImageCreationFailed(image_type='iso', error=e)


def fetch_into(context, image_href, image_file,
               image_auth_data=None):
    """Fetches image file contents into a file.

    :param context: A context object.
    :param image_href: The Image URL or reference to attempt to retrieve.
    :param image_file: The file handler or file name to write the requested
                       file contents to.
    :param image_auth_data: Optional dictionary for credentials to be conveyed
                            from the original task to the image download
                            process, if required.
    :returns: If a value is returned, that value was validated as the checksum.
              Otherwise None indicating the process had been completed.
    """
    # TODO(TheJulia): We likely need to document all of the exceptions which
    #                 can be raised by any of the various image services here.
    # TODO(vish): Improve context handling and add owner and auth data
    #             when it is added to glance.  Right now there is no
    #             auth checking in glance, so we assume that access was
    #             checked before we got here.
    image_service = service.get_image_service(image_href,
                                              context=context)
    LOG.debug("Using %(image_service)s to download image %(image_href)s.",
              {'image_service': image_service.__class__.__name__,
               'image_href': image_href})
    start = time.time()

    if image_service.is_auth_set_needed:
        # Send a dictionary with username/password data,
        # but send it in a dictionary since it fundimentally
        # can differ dramatically by types.
        image_service.set_image_auth(image_href, image_auth_data)

    if isinstance(image_file, str):
        with open(image_file, "wb") as image_file_obj:
            image_service.download(image_href, image_file_obj)
    else:
        image_service.download(image_href, image_file)

    LOG.debug("Image %(image_href)s downloaded in %(time).2f seconds.",
              {'image_href': image_href, 'time': time.time() - start})
    if image_service.transfer_verified_checksum:
        # TODO(TheJulia): The Glance Image service client does a
        # transfer related check when it retrieves the file. We might want
        # to shift the model some to do that upfront across most image
        # services which are able to be used that way.

        # We know, thanks to a value and not an exception, that
        # we have a checksum which matches the transfer.
        return image_service.transfer_verified_checksum
    return None


def _handle_zstd_compression(path):
    zstd_comp = False
    with open(path, 'rb') as comp_check:
        # Check for zstd compression. Zstd has a variable window for streaming
        # clients with transparent connections, and 128 byte blocks.
        # Ultimately, requests can't support handling of such content without
        # the zstandard library (bsd), but that is not available in global
        # requirements. As such, and likely best complexity wise, if we find it
        # we can handle it directly.
        # https://github.com/facebook/zstd/blob/dev/doc/zstd_compression_format.md
        # Ensure we're at the start of the file
        comp_check.seek(0)
        read = comp_check.read(4)
        if read.startswith(b"\x28\xb5\x2f\xfd"):
            zstd_comp = True

    if zstd_comp and not CONF.conductor.disable_zstandard_decompression:
        temp_path = path + '.zstd'
        shutil.move(path, temp_path)
        try:
            utils.execute('zstd', '-d', '--rm', temp_path)
        except OSError as e:
            LOG.error('Failed to decompress a zstd compressed file: %s', e)
            # Restore the downloaded file... We might want to fail the
            # entire process.
            shutil.move(temp_path, path)


def fetch(context, image_href, path, force_raw=False,
          checksum=None, checksum_algo=None,
          image_auth_data=None):
    with fileutils.remove_path_on_error(path):
        transfer_checksum = fetch_into(context, image_href, path,
                                       image_auth_data)
        if (not transfer_checksum
                and not CONF.conductor.disable_file_checksum
                and checksum):
            checksum_utils.validate_checksum(path, checksum, checksum_algo)

    # Check and decompress zstd files, since python-requests realistically
    # can't do it for us as-is. Also, some OCI container registry artifacts
    # may generally just be zstd compressed, regardless if it is a raw file
    # or a qcow2 file.
    _handle_zstd_compression(path)

    if force_raw:
        image_to_raw(image_href, path, "%s.part" % path)


def get_source_format(image_href, path):
    try:
        img_format = image_format_inspector.detect_file_format(path)
    except image_format_inspector.ImageFormatError as exc:
        LOG.error("Parsing of the image %s failed: %s", image_href, exc)
        raise exception.ImageUnacceptable(
            reason=_("parsing of the image failed."),
            image_id=image_href)
    if img_format is None:
        LOG.error("Parsing of the image %s failed: format not recognized",
                  image_href)
        raise exception.ImageUnacceptable(
            reason=_("parsing of the image failed."),
            image_id=image_href)
    return str(img_format)


def force_raw_will_convert(image_href, path_tmp):
    with fileutils.remove_path_on_error(path_tmp):
        fmt = get_source_format(image_href, path_tmp)
    return fmt not in RAW_IMAGE_FORMATS


def image_to_raw(image_href, path, path_tmp):
    with fileutils.remove_path_on_error(path_tmp):
        if not CONF.conductor.disable_deep_image_inspection:
            fmt = safety_check_image(path_tmp)

            if not image_format_permitted(fmt):
                LOG.error("Security: The requested image %(image_href)s "
                          "is of format image %(format)s and is not in "
                          "the [conductor]permitted_image_formats list.",
                          {'image_href': image_href,
                           'format': fmt})
                raise exception.InvalidImage()
        else:
            fmt = get_source_format(image_href, path)
            LOG.warning("Security: Image safety checking has been disabled. "
                        "This is unsafe operation. Attempting to continue "
                        "the detected format %(img_fmt)s for %(path)s.",
                        {'img_fmt': fmt,
                         'path': path})

        if fmt not in RAW_IMAGE_FORMATS and fmt != "iso":
            # When the target format is NOT raw, we need to convert it.
            # however, we don't need nor want to do that when we have
            # an ISO image. If we have an ISO because it was requested,
            # we have correctly fingerprinted it. Prior to proper
            # image detection, we thought we had a raw image, and we
            # would end up asking for a raw image to be made a raw image.
            staged = "%s.converted" % path

            utils.is_memory_insufficient(raise_if_fail=True)
            LOG.debug("%(image)s was %(format)s, converting to raw",
                      {'image': image_href, 'format': fmt})
            with fileutils.remove_path_on_error(staged):
                qemu_img.convert_image(path_tmp, staged, 'raw',
                                       source_format=fmt)
                os.unlink(path_tmp)
                new_fmt = get_source_format(image_href, staged)
                if new_fmt not in RAW_IMAGE_FORMATS:
                    raise exception.ImageConvertFailed(
                        image_id=image_href,
                        reason=_("Converted to raw, but format is "
                                 "now %s") % new_fmt)

                os.rename(staged, path)
        else:
            os.rename(path_tmp, path)


def image_show(context, image_href, image_service=None, image_auth_data=None):
    if image_service is None:
        image_service = service.get_image_service(image_href, context=context)
    if image_service.is_auth_set_needed:
        # We need to possibly authenticate, so we should attempt to do so.
        image_service.set_image_auth(image_href, image_auth_data)
    return image_service.show(image_href)


def download_size(context, image_href, image_service=None,
                  image_auth_data=None):
    return image_show(context, image_href,
                      image_service=image_service,
                      image_auth_data=image_auth_data)['size']


def converted_size(path, estimate=False):
    """Get size of converted raw image.

    The size of image converted to raw format can be growing up to the virtual
    size of the image.

    :param path: path to the image file.
    :param estimate: Whether to estimate the size by scaling the
        original size
    :returns: For `estimate=False`, return the size of the
        raw image file. For `estimate=True`, return the size of
        the original image scaled by the configuration value
        `raw_image_growth_factor`.
    """
    data = image_format_inspector.detect_file_format(path)
    if not estimate:
        return data.virtual_size
    growth_factor = CONF.raw_image_growth_factor
    return int(min(data.actual_size * growth_factor, data.virtual_size))


def get_image_properties(context, image_href, properties="all"):
    """Returns the values of several properties of an image

    :param context: context
    :param image_href: href of the image
    :param properties: the properties whose values are required.
        This argument is optional, default value is "all", so if not specified
        all properties will be returned.
    :returns: a dict of the values of the properties. A property not on the
        glance metadata will have a value of None.
    """
    img_service = service.get_image_service(image_href, context=context)
    iproperties = img_service.show(image_href)['properties']

    if properties == "all":
        return iproperties

    return {p: iproperties.get(p) for p in properties}


def get_temp_url_for_glance_image(context, image_uuid):
    """Returns the tmp url for a glance image.

    :param context: context
    :param image_uuid: the UUID of the image in glance
    :returns: the tmp url for the glance image.
    """
    glance_service = service.GlanceImageService(context=context)
    image_properties = glance_service.show(image_uuid)
    LOG.debug('Got image info: %(info)s for image %(image_uuid)s.',
              {'info': image_properties, 'image_uuid': image_uuid})
    return glance_service.swift_temp_url(image_properties)


def create_boot_iso(context, output_filename, kernel_href,
                    ramdisk_href, deploy_iso_href=None, esp_image_href=None,
                    root_uuid=None, kernel_params=None, boot_mode=None,
                    inject_files=None, publisher_id=None):
    """Creates a bootable ISO image for a node.

    Given the hrefs for kernel, ramdisk, root partition's UUID and
    kernel cmdline arguments, this method fetches the kernel and ramdisk,
    and builds a bootable ISO image that can be used to boot up the
    baremetal node.

    :param context: context
    :param output_filename: the absolute path of the output ISO file
    :param kernel_href: URL or glance uuid of the kernel to use
    :param ramdisk_href: URL or glance uuid of the ramdisk to use
    :param deploy_iso_href: URL or glance UUID of the deploy ISO image
        to extract EFI system partition image. If not specified,
        the `esp_image_href` option must be present if UEFI-bootable
        ISO is desired.
    :param esp_image_href: URL or glance UUID of FAT12/16/32-formatted EFI
        system partition image containing the EFI boot loader (e.g. GRUB2)
        for each hardware architecture to boot. This image will be written
        onto the ISO image. If not specified, the `deploy_iso_href` option
        is only required for building UEFI-bootable ISO.
    :param kernel_params: a string containing whitespace separated values
        kernel cmdline arguments of the form K=V or K (optional).
    :boot_mode: the boot mode in which the deploy is to happen.
    :param inject_files: Mapping of local source file paths to their location
        on the final ISO image.
    :param publisher_id: A value to set as the publisher identifier string
        in the ISO image to be generated.
    :raises: ImageCreationFailed, if creating boot ISO failed.
    """
    with utils.tempdir() as tmpdir:
        kernel_path = os.path.join(tmpdir, 'kernel')
        ramdisk_path = os.path.join(tmpdir, 'ramdisk')
        fetch(context, kernel_href, kernel_path)
        fetch(context, ramdisk_href, ramdisk_path)

        params = []
        if root_uuid:
            params.append('root=UUID=%s' % root_uuid)
        if kernel_params:
            params.append(kernel_params)

        if boot_mode == 'uefi':

            deploy_iso_path = esp_image_path = None

            if deploy_iso_href:
                deploy_iso_path = os.path.join(tmpdir, 'iso')
                fetch(context, deploy_iso_href, deploy_iso_path)

            elif esp_image_href:
                esp_image_path = os.path.join(tmpdir, 'esp')
                fetch(context, esp_image_href, esp_image_path)

            elif CONF.esp_image:
                esp_image_path = CONF.esp_image
            # TODO(TheJulia): we should opportunistically try to make bios
            # bootable and UEFI. In other words, collapse a lot of this
            # path since they are not mutually exclusive.
            # UEFI boot mode, but Network iPXE -> ISO means bios bootable
            # contents are still required.
            create_esp_image_for_uefi(
                output_filename, kernel_path, ramdisk_path,
                deploy_iso=deploy_iso_path, esp_image=esp_image_path,
                kernel_params=params, inject_files=inject_files,
                publisher_id=publisher_id)

        else:
            create_isolinux_image_for_bios(
                output_filename, kernel_path, ramdisk_path,
                kernel_params=params, inject_files=inject_files,
                publisher_id=publisher_id)


IMAGE_TYPE_PARTITION = 'partition'
IMAGE_TYPE_WHOLE_DISK = 'whole-disk'
VALID_IMAGE_TYPES = frozenset((IMAGE_TYPE_PARTITION, IMAGE_TYPE_WHOLE_DISK))


def is_whole_disk_image(ctx, instance_info):
    """Find out if the image is a partition image or a whole disk image.

    :param ctx: an admin context
    :param instance_info: a node's instance info dict

    :returns: True for whole disk images and False for partition images
        and None on no image_source, the source being a path, or upon an
        Error.
    """
    image_source = instance_info.get('image_source')
    if not image_source:
        return

    image_type = instance_info.get('image_type')
    if image_type:
        # This logic reflects the fact that whole disk images are the default
        return image_type != IMAGE_TYPE_PARTITION

    is_whole_disk_image = False
    if glance_utils.is_glance_image(image_source):
        try:
            iproperties = get_image_properties(ctx, image_source)
        except Exception:
            return

        image_type = iproperties.get('img_type')
        if image_type:
            return image_type != IMAGE_TYPE_PARTITION

        is_whole_disk_image = (not iproperties.get('kernel_id')
                               and not iproperties.get('ramdisk_id'))
    elif service.is_container_registry_url(image_source):
        # NOTE(theJulia): We can safely assume, at least outright,
        # that all container images are whole disk images, unelss
        # someone wants to add explicit support.
        is_whole_disk_image = True
    else:
        # Non glance image ref
        if is_source_a_path(ctx, instance_info.get('image_source')):
            # Nothing is returned if not valid or there was an error.
            # A third possibility is it is not a disk image, which would
            # still be None.
            return
        if (not instance_info.get('kernel')
            and not instance_info.get('ramdisk')):
            is_whole_disk_image = True

    return is_whole_disk_image


def is_source_a_path(ctx, image_source):
    """Determine if the image source is a path.

    This method determines if a supplied URL is a path.

    :param ctx: an admin/process context.
    :param image_source: The supplied image source, expected to be a
                         URL, which can be used to attempt to determine
                         if the source is a path.
    :returns: True if the image_source appears to be a path as opposed
              to an image to be downloaded. If the image source is not
              a path, False is returned. If any error is detected,
              None is returned.
    """
    if not image_source:
        return
    image_service = service.get_image_service(image_source,
                                              context=ctx)
    try:
        res = image_service.validate_href(image_source)
        if 'headers' in dir(res):
            # response/result is from the HTTP check path.
            headers = res.headers
        else:
            # We have no headers.
            headers = {}
    except exception.ImageRefIsARedirect as e:
        # Our exception handling formats this for us in this
        # case. \o/
        LOG.debug(str(e))
        # Servers redirect to a proper folder ending in a slash if
        # not supplied originally.
        if e.redirect_url and e.redirect_url.endswith('/'):
            return True
    except Exception:
        # NOTE(TheJulia): I don't really like this pattern, *but*
        # the wholedisk image support is similar.
        return
    # NOTE(TheJulia): Files should have been caught almost exclusively
    # before with the Content-Length check.
    # When the ISO is mounted and the webserver mount point url is
    # checked here, it has both 'Content-Length' and 'Content-Type'
    # due to which it always returns False. Hence switched the conditions.
    if ('Content-Type' in headers
            and str(headers['Content-Type']).startswith('text/html')):
        LOG.debug('Evaluated %(url)s to determine if it is a URL to a path '
                  'or a file. A Content-Type header was returned with a text '
                  'content, which suggests a file list was returned.',
                  {'url': image_source})
        return True
    # When issuing a head request, folders have no length
    # A list can be generated by the server.. This is a solid
    # hint.
    if ('Content-Type' in headers
            and (str(headers['Content-Type']) != 'text/html')
            and 'Content-Length' in headers):
        LOG.debug('Evaluated %(url)s to determine if it is a URL to a path '
                  'or a file. A Content-Length header was returned '
                  'suggesting file.',
                  {'url': image_source})
        # NOTE(TheJulia): Files on a webserver have a length which is returned
        # when headres are queried.
        return False
    if image_source.endswith('/'):
        # If all else fails, looks like a URL, and the server didn't give
        # us any hints.
        return True
    # We were unable to determine if this was a folder or a file.
    return False


def _extract_iso(extract_iso, extract_dir):
    # NOTE(rpittau): we could probably just extract the files we need
    # if we find them. Also we probably need to detect the correct iso
    # type (UDF, RR, JOLIET).
    iso = pycdlib.PyCdlib()
    iso.open(extract_iso)

    for dirname, dirlist, filelist in iso.walk(iso_path='/'):
        dir_path = dirname.lstrip('/')
        for dir_iso in dirlist:
            os.makedirs(os.path.join(extract_dir, dir_path, dir_iso))
        for file in filelist:
            file_path = os.path.join(extract_dir, dirname, file)
            iso.get_file_from_iso(
                os.path.join(extract_dir, dir_path, file),
                iso_path=file_path)

    iso.close()


def _get_deploy_iso_files(deploy_iso, mountdir):
    """This function opens up the deploy iso used for deploy.

    :param deploy_iso: path to the deploy iso where its
                       contents are fetched to.
    :raises: ImageCreationFailed if extraction fails.
    :returns: a tuple consisting of - 1. a dictionary containing
                                         the values as required
                                         by create_isolinux_image,
                                      2. efiboot.img relative path, and
                                      3. grub.cfg relative path.

    """
    e_img_rel_path = None
    e_img_path = None
    grub_rel_path = None
    grub_path = None

    try:
        _extract_iso(deploy_iso, mountdir)
    except Exception as e:
        LOG.exception("extracting the deploy iso failed.")
        raise exception.ImageCreationFailed(image_type='iso', error=e)

    try:
        for (dir, subdir, files) in os.walk(mountdir):
            if 'efiboot.img' in files:
                e_img_path = os.path.join(dir, 'efiboot.img')
                e_img_rel_path = os.path.relpath(e_img_path,
                                                 mountdir)
            if 'grub.cfg' in files:
                grub_path = os.path.join(dir, 'grub.cfg')
                grub_rel_path = os.path.relpath(grub_path,
                                                mountdir)
    except (OSError, IOError) as e:
        LOG.exception("examining the deploy iso failed.")
        shutil.rmtree(mountdir)
        raise exception.ImageCreationFailed(image_type='iso', error=e)

    # check if the variables are assigned some values or not during
    # walk of the mountdir.
    if not (e_img_path and e_img_rel_path and grub_path and grub_rel_path):
        error = (_("Deploy iso didn't contain efiboot.img or grub.cfg"))
        shutil.rmtree(mountdir)
        raise exception.ImageCreationFailed(image_type='iso', error=error)

    uefi_path_info = {e_img_path: e_img_rel_path,
                      grub_path: grub_rel_path}

    # Returning a tuple as it makes the code simpler and clean.
    # uefi_path_info: is needed by the caller for _create_root_fs to create
    # appropriate directory structures for uefi boot iso.
    # grub_rel_path: is needed to copy the new grub.cfg generated using
    # generate_cfg() to the same directory path structure where it was
    # present in deploy iso. This path varies for different OS vendors.
    # e_img_rel_path: is required by mkisofs to generate boot iso.
    return uefi_path_info, e_img_rel_path, grub_rel_path


def __node_or_image_cache(node):
    """A helper for logging to determine if image cache or node uuid."""
    if not node:
        return 'image cache'
    else:
        return node.uuid


def safety_check_image(image_path, node=None):
    """Performs a safety check on the supplied image.

    This method triggers the image format inspector's to both identify the
    type of the supplied file and safety check logic to identify if there
    are any known unsafe features being leveraged, and return the detected
    file format in the form of a string for the caller.

    :param image_path: A fully qualified path to an image which needs to
                       be evaluated for safety.
    :param node: A Node object, optional. When supplied logging indicates the
                 node which triggered this issue, but the node is not
                 available in all invocation cases.
    :returns: a string representing the the image type which is used.
    :raises: InvalidImage when the supplied image is detected as unsafe,
             or the image format inspector has failed to parse the supplied
             image's contents.
    """
    id_string = __node_or_image_cache(node)
    try:
        img_class = image_format_inspector.detect_file_format(image_path)
        if img_class is None:
            LOG.error("Security: The requested user image for the "
                      "deployment node %(node)s does not match any known "
                      "format",
                      {'node': id_string})
            raise exception.InvalidImage()
        img_class.safety_check()
        image_format_name = str(img_class)
    except image_format_inspector.ImageFormatError as exc:
        LOG.error("Security: The requested user image for the "
                  "deployment node %(node)s failed to be able "
                  "to be parsed by the image format checker: %(exc)s",
                  {'node': id_string, 'exc': exc})
        raise exception.InvalidImage()
    except image_format_inspector.SafetyCheckFailed as exc:
        LOG.error("Security: The requested image for "
                  "deployment of node %(node)s fails safety sanity "
                  "checking: %(exc)s",
                  {'node': id_string, 'exc': exc})
        raise exception.InvalidImage()
    return image_format_name


RAW_IMAGE_FORMATS = {'raw', 'gpt'}  # gpt is a whole-disk image


def image_format_permitted(img_format):
    permitted = set(CONF.conductor.permitted_image_formats)
    if 'raw' in permitted:
        permitted.update(RAW_IMAGE_FORMATS)
    return img_format in permitted


def image_format_matches(actual_format, expected_format):
    if expected_format in ['ari', 'aki']:
        # In this case, we have an ari or aki, meaning we're pulling
        # down a kernel/ramdisk, and this is rooted in a misunderstanding.
        # They should be raw. The detector should be detecting this *as*
        # raw anyway, so the data just mismatches from a common
        # misunderstanding, and that is okay in this case as they are not
        # passed to qemu-img.
        # TODO(TheJulia): Add a log entry to warn here at some point in
        # the future as we begin to shift the perception around this.
        # See: https://bugs.launchpad.net/ironic/+bug/2074090
        return True
    if expected_format == 'raw' and actual_format in RAW_IMAGE_FORMATS:
        return True
    return expected_format == actual_format


def check_if_image_format_is_permitted(img_format,
                                       expected_format=None,
                                       node=None):
    """Checks image format consistency.

    :params img_format: The determined image format by name.
    :params expected_format: Optional, the expected format based upon
        supplied configuration values.
    :params node: A node object or None implying image cache.
    :raises: InvalidImage if the requested image format is not permitted
             by configuration, or the expected_format does not match the
             determined format.
    """

    id_string = __node_or_image_cache(node)
    if not image_format_permitted(img_format):
        LOG.error("Security: The requested deploy image for node %(node)s "
                  "is of format image %(format)s and is not in the "
                  "[conductor]permitted_image_formats list.",
                  {'node': id_string,
                   'format': img_format})
        raise exception.InvalidImage()
    if (expected_format is not None
            and not image_format_matches(img_format, expected_format)):
        LOG.error("Security: The requested deploy image for node %(node)s "
                  "has a format (%(format)s) which does not match the "
                  "expected image format (%(expected)s) based upon "
                  "supplied or retrieved information.",
                  {'node': id_string,
                   'format': img_format,
                   'expected': expected_format})
        raise exception.InvalidImage()



# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from webob import exc as webob_exc

from ironic.api.controllers import v1 as v1_api
from ironic.api.controllers.v1 import versions
from ironic.tests import base as test_base
from ironic.tests.unit.api import base as api_base


class TestV1Routing(api_base.BaseApiTest):
    def test_route_checks_version(self):
        self.get_json('/')
        self._check_version.assert_called_once_with(mock.ANY,
                                                    mock.ANY,
                                                    mock.ANY)

    def test_microversion_headers(self):
        # Check root endpoint
        response = self.app.head('/')
        self.assertEqual(200, response.status_int)
        self.assertIn('X-OpenStack-Ironic-API-Minimum-Version',
                      response.headers)
        self.assertIn('X-OpenStack-Ironic-API-Maximum-Version',
                      response.headers)

        self._check_version.reset_mock()

        # Check v1 endpoint
        response = self.app.head('/v1')
        self.assertEqual(200, response.status_int)
        self.assertIn('X-OpenStack-Ironic-API-Minimum-Version',
                      response.headers)
        self.assertIn('X-OpenStack-Ironic-API-Maximum-Version',
                      response.headers)

        # -check_version should be called only once
        self.assertEqual(1, self._check_version.call_count)

    def test_min_version(self):
        response = self.get_json(
            '/',
            headers={
                'Accept': 'application/json',
                'X-OpenStack-Ironic-API-Version':
                versions.min_version_string()
            })
        self.assertEqual({
            'id': 'v1',
            'links': [
                {'href': 'http://localhost/v1/', 'rel': 'self'},
                {'href': 'https://docs.openstack.org//ironic/latest'
                         '/contributor//webapi.html',
                 'rel': 'describedby', 'type': 'text/html'}
            ],
            'media_types': {
                'base': 'application/json',
                'type': 'application/vnd.openstack.ironic.v1+json'
            },
            'version': {
                'id': 'v1',
                'links': [{'href': 'http://localhost/v1/', 'rel': 'self'}],
                'status': 'CURRENT',
                'min_version': versions.min_version_string(),
                'version': versions.max_version_string()
            },
            'chassis': [
                {'href': 'http://localhost/v1/chassis/', 'rel': 'self'},
                {'href': 'http://localhost/chassis/', 'rel': 'bookmark'}
            ],
            'nodes': [
                {'href': 'http://localhost/v1/nodes/', 'rel': 'self'},
                {'href': 'http://localhost/nodes/', 'rel': 'bookmark'}
            ],
            'ports': [
                {'href': 'http://localhost/v1/ports/', 'rel': 'self'},
                {'href': 'http://localhost/ports/', 'rel': 'bookmark'}
            ],
            'drivers': [
                {'href': 'http://localhost/v1/drivers/', 'rel': 'self'},
                {'href': 'http://localhost/drivers/', 'rel': 'bookmark'}
            ],
        }, response)

    def test_max_version(self):
        response = self.get_json(
            '/',
            headers={
                'Accept': 'application/json',
                'X-OpenStack-Ironic-API-Version':
                versions.max_version_string()
            })
        self.assertEqual({
            'id': 'v1',
            'inspection_rules': [
                {'href': 'http://localhost/v1/inspection_rules/',
                 'rel': 'self'},
                {'href': 'http://localhost/inspection_rules/',
                 'rel': 'bookmark'}
            ],
            'links': [
                {'href': 'http://localhost/v1/', 'rel': 'self'},
                {'href': 'https://docs.openstack.org//ironic/latest'
                         '/contributor//webapi.html',
                 'rel': 'describedby', 'type': 'text/html'}
            ],
            'media_types': {
                'base': 'application/json',
                'type': 'application/vnd.openstack.ironic.v1+json'
            },
            'version': {
                'id': 'v1',
                'links': [{'href': 'http://localhost/v1/', 'rel': 'self'}],
                'status': 'CURRENT',
                'min_version': versions.min_version_string(),
                'version': versions.max_version_string()
            },
            'allocations': [
                {'href': 'http://localhost/v1/allocations/', 'rel': 'self'},
                {'href': 'http://localhost/allocations/', 'rel': 'bookmark'}
            ],
            'chassis': [
                {'href': 'http://localhost/v1/chassis/', 'rel': 'self'},
                {'href': 'http://localhost/chassis/', 'rel': 'bookmark'}
            ],
            'conductors': [
                {'href': 'http://localhost/v1/conductors/', 'rel': 'self'},
                {'href': 'http://localhost/conductors/', 'rel': 'bookmark'}
            ],
            'continue_inspection': [
                {'href': 'http://localhost/v1/continue_inspection/',
                 'rel': 'self'},
                {'href': 'http://localhost/continue_inspection/',
                 'rel': 'bookmark'}
            ],
            'deploy_templates': [
                {'href': 'http://localhost/v1/deploy_templates/',
                 'rel': 'self'},
                {'href': 'http://localhost/deploy_templates/',
                 'rel': 'bookmark'}
            ],
            'drivers': [
                {'href': 'http://localhost/v1/drivers/', 'rel': 'self'},
                {'href': 'http://localhost/drivers/', 'rel': 'bookmark'}
            ],
            'events': [
                {'href': 'http://localhost/v1/events/', 'rel': 'self'},
                {'href': 'http://localhost/events/', 'rel': 'bookmark'}
            ],
            'heartbeat': [
                {'href': 'http://localhost/v1/heartbeat/', 'rel': 'self'},
                {'href': 'http://localhost/heartbeat/', 'rel': 'bookmark'}
            ],
            'lookup': [
                {'href': 'http://localhost/v1/lookup/', 'rel': 'self'},
                {'href': 'http://localhost/lookup/', 'rel': 'bookmark'}
            ],
            'nodes': [
                {'href': 'http://localhost/v1/nodes/', 'rel': 'self'},
                {'href': 'http://localhost/nodes/', 'rel': 'bookmark'}
            ],
            'portgroups': [
                {'href': 'http://localhost/v1/portgroups/', 'rel': 'self'},
                {'href': 'http://localhost/portgroups/', 'rel': 'bookmark'}
            ],
            'ports': [
                {'href': 'http://localhost/v1/ports/', 'rel': 'self'},
                {'href': 'http://localhost/ports/', 'rel': 'bookmark'}
            ],
            'shards': [
                {'href': 'http://localhost/v1/shards/', 'rel': 'self'},
                {'href': 'http://localhost/shards/', 'rel': 'bookmark'}
            ],
            'volume': [
                {'href': 'http://localhost/v1/volume/', 'rel': 'self'},
                {'href': 'http://localhost/volume/', 'rel': 'bookmark'}
            ],
            'runbooks': [
                {'href': 'http://localhost/v1/runbooks/',
                 'rel': 'self'},
                {'href': 'http://localhost/runbooks/',
                 'rel': 'bookmark'}
            ]
        }, response)


class TestCheckVersions(test_base.TestCase):

    def setUp(self):
        super(TestCheckVersions, self).setUp()

        class ver(object):
            major = None
            minor = None

        self.version = ver()

    def test_check_version_invalid_major_version(self):
        self.version.major = v1_api.BASE_VERSION + 1
        self.version.minor = v1_api.min_version().minor
        self.assertRaises(
            webob_exc.HTTPNotAcceptable,
            v1_api.Controller()._check_version,
            self.version)

    def test_check_version_too_low(self):
        self.version.major = v1_api.BASE_VERSION
        self.version.minor = v1_api.min_version().minor - 1
        self.assertRaises(
            webob_exc.HTTPNotAcceptable,
            v1_api.Controller()._check_version,
            self.version)

    def test_check_version_too_high(self):
        self.version.major = v1_api.BASE_VERSION
        self.version.minor = v1_api.max_version().minor + 1
        e = self.assertRaises(
            webob_exc.HTTPNotAcceptable,
            v1_api.Controller()._check_version,
            self.version, {'fake-headers': v1_api.max_version().minor})
        self.assertEqual(v1_api.max_version().minor, e.headers['fake-headers'])

    def test_check_version_ok(self):
        self.version.major = v1_api.BASE_VERSION
        self.version.minor = v1_api.min_version().minor
        v1_api.Controller()._check_version(self.version)
